import math
from collections.abc import Collection
from typing import Literal

import dask.array as da
import numpy as np
import zarr

from ngio.common._zoom import _zoom_inputs_check, dask_zoom, numpy_zoom
from ngio.utils import (
    AccessModeLiteral,
    NgioValueError,
    StoreOrGroup,
    open_group_wrapper,
)


def _on_disk_numpy_zoom(
    source: zarr.Array,
    target: zarr.Array,
    order: Literal[0, 1, 2] = 1,
) -> None:
    target[...] = numpy_zoom(source[...], target_shape=target.shape, order=order)


def _on_disk_dask_zoom(
    source: zarr.Array,
    target: zarr.Array,
    order: Literal[0, 1, 2] = 1,
) -> None:
    source_array = da.from_zarr(source)
    target_array = dask_zoom(source_array, target_shape=target.shape, order=order)

    target_array = target_array.rechunk(target.chunks)
    target_array.compute_chunk_sizes()
    target_array.to_zarr(target)


def _on_disk_coarsen(
    source: zarr.Array,
    target: zarr.Array,
    _order: Literal[0, 1] = 1,
    aggregation_function: np.ufunc | None = None,
) -> None:
    """Apply a coarsening operation from a source zarr array to a target zarr array.

    Args:
        source (zarr.Array): The source array to coarsen.
        target (zarr.Array): The target array to save the coarsened result to.
        _order (Literal[0, 1]): The order of interpolation is not really implemented
            for coarsening, but it is kept for compatibility with the zoom function.
            _order=1 -> linear interpolation ~ np.mean
            _order=0 -> nearest interpolation ~ np.max
        aggregation_function (np.ufunc): The aggregation function to use.
    """
    source_array = da.from_zarr(source)

    _scale, _target_shape = _zoom_inputs_check(
        source_array=source_array, scale=None, target_shape=target.shape
    )

    assert _target_shape == target.shape, (
        "Target shape must match the target array shape"
    )

    if aggregation_function is None:
        if _order == 1:
            aggregation_function = np.mean
        elif _order == 0:
            aggregation_function = np.max
        else:
            raise NgioValueError(
                f"Aggregation function must be provided for order {_order}"
            )

    coarsening_setup = {}
    for i, s in enumerate(_scale):
        factor = 1 / s
        # This check is very strict, but it is necessary to avoid
        # a few pixels shift in the coarsening
        # We could add a tolerance
        if factor.is_integer():
            coarsening_setup[i] = int(factor)
        else:
            raise NgioValueError(
                f"Coarsening factor must be an integer, got {factor} on axis {i}"
            )

    out_target = da.coarsen(
        aggregation_function, source_array, coarsening_setup, trim_excess=True
    )
    out_target = out_target.rechunk(target.chunks)
    out_target.to_zarr(target)


def on_disk_zoom(
    source: zarr.Array,
    target: zarr.Array,
    order: Literal[0, 1, 2] = 1,
    mode: Literal["dask", "numpy", "coarsen"] = "dask",
) -> None:
    """Apply a zoom operation from a source zarr array to a target zarr array.

    Args:
        source (zarr.Array): The source array to zoom.
        target (zarr.Array): The target array to save the zoomed result to.
        order (Literal[0, 1, 2]): The order of interpolation. Defaults to 1.
        mode (Literal["dask", "numpy", "coarsen"]): The mode to use. Defaults to "dask".
    """
    if not isinstance(source, zarr.Array):
        raise NgioValueError("source must be a zarr array")

    if not isinstance(target, zarr.Array):
        raise NgioValueError("target must be a zarr array")

    if source.dtype != target.dtype:
        raise NgioValueError("source and target must have the same dtype")

    match mode:
        case "numpy":
            return _on_disk_numpy_zoom(source, target, order)
        case "dask":
            return _on_disk_dask_zoom(source, target, order)
        case "coarsen":
            return _on_disk_coarsen(
                source,
                target,
            )
        case _:
            raise NgioValueError("mode must be either 'dask', 'numpy' or 'coarsen'")


def _find_closest_arrays(
    processed: list[zarr.Array], to_be_processed: list[zarr.Array]
) -> tuple[int, int]:
    dist_matrix = np.zeros((len(processed), len(to_be_processed)))
    for i, arr_to_proc in enumerate(to_be_processed):
        for j, proc_arr in enumerate(processed):
            dist_matrix[j, i] = np.sqrt(
                np.sum(
                    [
                        (s1 - s2) ** 2
                        for s1, s2 in zip(
                            arr_to_proc.shape, proc_arr.shape, strict=False
                        )
                    ]
                )
            )

    return np.unravel_index(dist_matrix.argmin(), dist_matrix.shape)


def consolidate_pyramid(
    source: zarr.Array,
    targets: list[zarr.Array],
    order: Literal[0, 1, 2] = 1,
    mode: Literal["dask", "numpy", "coarsen"] = "dask",
) -> None:
    """Consolidate the Zarr array."""
    processed = [source]
    to_be_processed = targets

    while to_be_processed:
        source_id, target_id = _find_closest_arrays(processed, to_be_processed)

        source_image = processed[source_id]
        target_image = to_be_processed.pop(target_id)

        on_disk_zoom(
            source=source_image,
            target=target_image,
            mode=mode,
            order=order,
        )
        processed.append(target_image)


def init_empty_pyramid(
    store: StoreOrGroup,
    paths: list[str],
    ref_shape: Collection[int],
    scaling_factors: Collection[float],
    chunks: Collection[int] | None = None,
    dtype: str = "uint16",
    mode: AccessModeLiteral = "a",
) -> None:
    # Return the an Image object
    if chunks is not None and len(chunks) != len(ref_shape):
        raise NgioValueError(
            "The shape and chunks must have the same number of dimensions."
        )

    if chunks is not None:
        chunks = [min(c, s) for c, s in zip(chunks, ref_shape, strict=True)]

    if len(ref_shape) != len(scaling_factors):
        raise NgioValueError(
            "The shape and scaling factor must have the same number of dimensions."
        )

    root_group = open_group_wrapper(store, mode=mode)

    for path in paths:
        if any(s < 1 for s in ref_shape):
            raise NgioValueError(
                "Level shape must be at least 1 on all dimensions. "
                f"Calculated shape: {ref_shape} at level {path}."
            )
        new_arr = root_group.zeros(
            name=path,
            shape=ref_shape,
            dtype=dtype,
            chunks=chunks,
            dimension_separator="/",
            overwrite=True,
        )

        # Todo redo this with when a proper build of pyramid is implemented
        _shape = []
        for s, sc in zip(ref_shape, scaling_factors, strict=True):
            if math.floor(s / sc) % 2 == 0:
                _shape.append(math.floor(s / sc))
            else:
                _shape.append(math.ceil(s / sc))
        ref_shape = _shape

        if chunks is None:
            chunks = new_arr.chunks
            if chunks is None:
                raise NgioValueError("Something went wrong with the chunks")
        chunks = [min(c, s) for c, s in zip(chunks, ref_shape, strict=True)]
    return None



"""Common classes and functions that are used across the package."""

from ngio.common._array_pipe import (
    get_masked_pipe,
    get_pipe,
    set_masked_pipe,
    set_pipe,
)
from ngio.common._axes_transforms import (
    transform_dask_array,
    transform_list,
    transform_numpy_array,
)
from ngio.common._common_types import ArrayLike
from ngio.common._dimensions import Dimensions
from ngio.common._masking_roi import compute_masking_roi
from ngio.common._pyramid import consolidate_pyramid, init_empty_pyramid, on_disk_zoom
from ngio.common._roi import Roi, RoiPixels, roi_to_slice_kwargs
from ngio.common._slicer import (
    SliceTransform,
    compute_and_slices,
    dask_get_slice,
    dask_set_slice,
    numpy_get_slice,
    numpy_set_slice,
)
from ngio.common._zoom import dask_zoom, numpy_zoom

__all__ = [
    "ArrayLike",
    "Dimensions",
    "Roi",
    "RoiPixels",
    "SliceTransform",
    "compute_and_slices",
    "compute_masking_roi",
    "consolidate_pyramid",
    "dask_get_slice",
    "dask_set_slice",
    "dask_zoom",
    "get_masked_pipe",
    "get_pipe",
    "init_empty_pyramid",
    "numpy_get_slice",
    "numpy_set_slice",
    "numpy_zoom",
    "on_disk_zoom",
    "roi_to_slice_kwargs",
    "set_masked_pipe",
    "set_pipe",
    "transform_dask_array",
    "transform_list",
    "transform_numpy_array",
]



from abc import ABC, abstractmethod
from typing import Literal

from anndata import AnnData
from pandas import DataFrame
from polars import DataFrame as PolarsDataFrame
from polars import LazyFrame
from pydantic import BaseModel

from ngio.tables.backends._utils import (
    convert_anndata_to_pandas,
    convert_anndata_to_polars,
    convert_pandas_to_anndata,
    convert_pandas_to_polars,
    convert_polars_to_anndata,
    convert_polars_to_pandas,
)
from ngio.utils import NgioValueError, ZarrGroupHandler


class BackendMeta(BaseModel):
    """Metadata for the backend."""

    backend: str | None = None
    index_key: str | None = None
    index_type: Literal["int", "str"] | None = None


class AbstractTableBackend(ABC):
    """Abstract class for table backends."""

    def __init__(
        self,
        group_handler: ZarrGroupHandler,
        index_key: str | None = None,
        index_type: Literal["int", "str"] | None = None,
    ):
        """Initialize the handler.

        This is a base class for the table backends protocol.

        Args:
            group_handler (ZarrGroupHandler): An object to handle the Zarr group
                containing the table data.
            index_key (str): The column name to use as the index of the DataFrame.
            index_type (str): The type of the index column in the DataFrame.
        """
        self._group_handler = group_handler
        self._index_key = index_key
        self._index_type = index_type

    @staticmethod
    @abstractmethod
    def backend_name() -> str:
        """Return the name of the backend.

        As a convention we set name as:
            {backend_name}_v{version}

        Where the version is a integer.
        """
        raise NotImplementedError

    @staticmethod
    @abstractmethod
    def implements_anndata() -> bool:
        """Check if the backend implements the anndata protocol.

        If this is True, the backend should implement the
        `load_as_anndata` and `write_from_anndata` methods.

        If this is False, these methods should raise a
        `NotImplementedError`.
        """
        raise NotImplementedError

    @staticmethod
    @abstractmethod
    def implements_pandas() -> bool:
        """Check if the backend implements the pandas protocol.

        If this is True, the backend should implement the
        `load_as_dataframe` and `write_from_dataframe` methods.

        If this is False, these methods should raise a
        `NotImplementedError`.
        """
        raise NotImplementedError

    @staticmethod
    @abstractmethod
    def implements_polars() -> bool:
        """Check if the backend implements the polars protocol.

        If this is True, the backend should implement the
        `load_as_polars` and `write_from_polars` methods.

        If this is False, these methods should raise a
        `NotImplementedError`.
        """
        raise NotImplementedError

    @property
    def group_handler(self) -> ZarrGroupHandler:
        """Get the group handler."""
        return self._group_handler

    @property
    def index_key(self) -> str | None:
        """Get the index key."""
        return self._index_key

    @property
    def index_type(self) -> Literal["int", "str"] | None:
        """Get the index type."""
        if self._index_type is None:
            return None

        if self._index_type not in ["int", "str"]:
            raise NgioValueError(
                f"Invalid index type {self._index_type}. Must be 'int' or 'str'."
            )
        return self._index_type  # type: ignore[return-value]

    def load_as_anndata(self) -> AnnData:
        """Load the table as an AnnData object.

        Since the AnnData object is more complex than a DataFrame,
        selecting columns is not implemented, because it is not
        straightforward to do so for an arbitrary AnnData object.
        """
        if self.implements_pandas():
            return convert_pandas_to_anndata(
                self.load_as_pandas_df(),
                index_key=self.index_key,
            )
        elif self.implements_polars():
            return convert_polars_to_anndata(
                self.load_as_polars_lf(),
                index_key=self.index_key,
            )
        else:
            raise NgioValueError(
                "Backend does not implement any of the protocols. "
                "A backend should implement at least one of the "
                "following protocols: anndata, pandas, polars."
            )

    def load_as_pandas_df(self) -> DataFrame:
        """Load the table as a pandas DataFrame.

        If columns are provided, the table should be filtered
        """
        if self.implements_anndata():
            return convert_anndata_to_pandas(
                self.load_as_anndata(),
                index_key=self.index_key,
                index_type=self.index_type,
            )
        elif self.implements_polars():
            return convert_polars_to_pandas(
                self.load_as_polars_lf(),
                index_key=self.index_key,
                index_type=self.index_type,
            )
        else:
            raise NgioValueError(
                "Backend does not implement any of the protocols. "
                "A backend should implement at least one of the "
                "following protocols: anndata, pandas, polars."
            )

    def load_as_polars_lf(self) -> LazyFrame:
        """Load the table as a polars LazyFrame.

        If columns are provided, the table should be filtered
        """
        if self.implements_anndata():
            return convert_anndata_to_polars(
                self.load_as_anndata(),
                index_key=self.index_key,
                index_type=self.index_type,
            ).lazy()
        elif self.implements_pandas():
            return convert_pandas_to_polars(
                self.load_as_pandas_df(),
                index_key=self.index_key,
                index_type=self.index_type,
            ).lazy()
        else:
            raise NgioValueError(
                "Backend does not implement any of the protocols. "
                "A backend should implement at least one of the "
                "following protocols: anndata, pandas, polars."
            )

    def write_from_pandas(self, table: DataFrame) -> None:
        """Serialize the table from a pandas DataFrame."""
        raise NotImplementedError(
            f"Backend {self.backend_name()} does not support "
            "serialization of DataFrame objects."
        )

    def write_from_anndata(self, table: AnnData) -> None:
        """Serialize the table from an AnnData object."""
        raise NotImplementedError(
            f"Backend {self.backend_name()} does not support "
            "serialization of AnnData objects."
        )

    def write_from_polars(self, table: PolarsDataFrame | LazyFrame) -> None:
        """Serialize the table from a polars DataFrame or LazyFrame."""
        raise NotImplementedError(
            f"Backend {self.backend_name()} does not support "
            "serialization of Polars objects."
        )

    def write_metadata(self, metadata: dict | None = None) -> None:
        """Write the metadata to the store."""
        if metadata is None:
            metadata = {}

        backend_metadata = BackendMeta(
            backend=self.backend_name(),
            index_key=self.index_key,
            index_type=self.index_type,
        ).model_dump(exclude_none=True)
        metadata.update(backend_metadata)
        self._group_handler.write_attrs(metadata)

    def write(
        self,
        table: DataFrame | AnnData | PolarsDataFrame | LazyFrame,
        metadata: dict | None = None,
        mode: Literal["pandas", "anndata", "polars"] | None = None,
    ) -> None:
        """Serialize the table to the store, and write the metadata.

        This is a generic write method.
        Based on the explicit mode or the type of the table,
        it will call the appropriate write method.
        """
        if mode == "pandas" or isinstance(table, DataFrame):
            self.write_from_pandas(table)  # type: ignore[arg-type]
        elif mode == "anndata" or isinstance(table, AnnData):
            self.write_from_anndata(table)  # type: ignore[arg-type]
        elif mode == "polars" or isinstance(table, PolarsDataFrame | LazyFrame):
            self.write_from_polars(table)
        else:
            raise NgioValueError(
                f"Unsupported table type {type(table)}. "
                "Please specify the mode explicitly. "
                "Supported serialization modes are: "
                "'pandas', 'anndata', 'polars'."
            )
        self.write_metadata(metadata)



"""Test suite for the utils module."""

import numpy as np
import pandas as pd
import pandas.testing as pdt
import polars as pl
import polars.testing as pl_testing
import pytest
from anndata import AnnData

from ngio.tables.backends._utils import (
    convert_anndata_to_pandas,
    convert_anndata_to_polars,
    convert_pandas_to_anndata,
    convert_pandas_to_polars,
    convert_polars_to_anndata,
    convert_polars_to_pandas,
    normalize_anndata,
    normalize_pandas_df,
    normalize_polars_lf,
)
from ngio.utils import NgioTableValidationError, NgioValueError


def sample_pandas_df_no_index():
    """Create a sample pandas DataFrame for testing."""
    return pd.DataFrame(
        {
            "id": ["1", "2", "3"],
            "str_col": ["x", "y", "z"],
            "int_col": [1, 2, 3],
            "float_col": [1.1, 2.2, 3.3],
        }
    )


def sample_pandas_df_int_index():
    """Create a sample pandas DataFrame for testing."""
    return pd.DataFrame(
        {
            "id": [1, 2, 3],
            "str_col": ["x", "y", "z"],
            "int_col": [1, 2, 3],
            "float_col": [1.1, 2.2, 3.3],
        }
    ).set_index("id")


def sample_pandas_df_str_index():
    """Create a sample pandas DataFrame for testing."""
    return pd.DataFrame(
        {
            "id": ["1", "2", "3"],
            "str_col": ["x", "y", "z"],
            "int_col": [1, 2, 3],
            "float_col": [1.1, 2.2, 3.3],
        }
    ).set_index("id")


def sample_polars_lf():
    """Create a sample polars DataFrame for testing."""
    return pl.DataFrame(
        {
            "id": ["1", "2", "3"],
            "str_col": ["x", "y", "z"],
            "int_col": [1, 2, 3],
            "float_col": [1.1, 2.2, 3.3],
        }
    ).lazy()


def sample_anndata():
    """Create a sample AnnData object for testing."""
    obs = pd.DataFrame(
        {
            "id": ["1", "2", "3"],
            "str_col": ["x", "y", "z"],
            "int_col": [1, 2, 3],
        }
    ).set_index("id")

    x = np.array([[1.1, 2.2, 3.2]]).T
    return AnnData(X=x, obs=obs)


@pytest.mark.parametrize(
    "in_df, out_df, index_key, index_type, reset_index",
    [
        (sample_pandas_df_no_index(), sample_pandas_df_no_index(), None, None, False),
        (sample_pandas_df_no_index(), sample_pandas_df_str_index(), "id", "str", False),
        (sample_pandas_df_no_index(), sample_pandas_df_int_index(), "id", "int", False),
        (sample_pandas_df_int_index(), sample_pandas_df_int_index(), None, None, False),
        (
            sample_pandas_df_int_index(),
            sample_pandas_df_str_index(),
            "id",
            "str",
            False,
        ),
        (sample_pandas_df_int_index(), sample_pandas_df_no_index(), "id", "str", True),
        (sample_pandas_df_str_index(), sample_pandas_df_no_index(), "id", "str", True),
    ],
)
def test_normalize_pandas_df(in_df, out_df, index_key, index_type, reset_index):
    """Test the normalization of a pandas DataFrame."""
    df = normalize_pandas_df(
        in_df,
        index_key=index_key,
        index_type=index_type,
        reset_index=reset_index,
    )

    pdt.assert_frame_equal(
        df,
        out_df,
        check_dtype=False,
    )


def test_normalize_pandas_df_index_none():
    """Test the normalization of a pandas DataFrame."""
    df = normalize_pandas_df(
        sample_pandas_df_no_index(),
        index_key="new_index_name",
        index_type="str",
        reset_index=False,
    )
    assert df.index.name == "new_index_name"


def test_fail_normalize_pandas_df():
    """Test the normalization of a pandas DataFrame."""
    with pytest.raises(NgioTableValidationError):
        normalize_pandas_df(
            sample_pandas_df_no_index(),
            index_key="str_col",
            index_type="int",
            reset_index=False,
        )

    with pytest.raises(NgioTableValidationError):
        normalize_pandas_df(
            sample_pandas_df_str_index(),
            index_key="not_exist",
            index_type="str",
            reset_index=False,
        )

    with pytest.raises(NgioValueError):
        normalize_pandas_df(
            sample_pandas_df_no_index(),
            index_key="id",
            index_type="float",  # type: ignore
            reset_index=False,
        )


def test_normalize_polars_lf():
    """Test the conversion of an AnnData object to a pandas DataFrame."""
    lf = sample_polars_lf()
    _ = normalize_polars_lf(lf, index_key="id", index_type="str").collect()
    _ = normalize_polars_lf(lf, index_key="id", index_type="int").collect()


def test_normalize_anndata():
    """Test the conversion of an AnnData object to a pandas DataFrame."""
    adata = sample_anndata()
    _ = normalize_anndata(adata, index_key="str_col")
    _ = normalize_anndata(adata, index_key="id")

    with pytest.raises(NgioTableValidationError):
        normalize_anndata(adata, index_key="not_exist")


def test_convert_pandas_to_anndata_roundtrip():
    """Test the conversion of an AnnData object to a pandas DataFrame."""
    df = sample_pandas_df_no_index()
    adata = convert_pandas_to_anndata(df, index_key="id")
    df_back = convert_anndata_to_pandas(adata, index_key="id", reset_index=True)

    for column in df.columns:
        pdt.assert_series_equal(df[column], df_back[column], check_index=False)


def test_convert_pandas_to_polars_roundtrip():
    """Test the conversion of a pandas DataFrame to an AnnData object."""
    df = sample_pandas_df_no_index()
    lf = convert_pandas_to_polars(df, index_key="id")
    df_back = convert_polars_to_pandas(lf, index_key="id", reset_index=True)

    for column in df.columns:
        pdt.assert_series_equal(df[column], df_back[column], check_index=False)


def test_other_conversions():
    """Test the conversion of a polars DataFrame to an AnnData object."""
    lf = sample_polars_lf().collect()
    adata = convert_polars_to_anndata(lf, index_key="id")
    lf_back = convert_anndata_to_polars(adata, index_key="id").collect()
    pl_testing.assert_frame_equal(
        lf,
        lf_back,
        check_column_order=False,
    )



from typing import Annotated

from ome_zarr_models.v04.well import WellAttrs as WellAttrs04
from ome_zarr_models.v04.well_types import WellImage as WellImage04
from ome_zarr_models.v04.well_types import WellMeta as WellMeta04
from pydantic import SkipValidation


class CustomWellImage(WellImage04):
    path: Annotated[str, SkipValidation]


class CustomWellMeta(WellMeta04):
    images: list[CustomWellImage]  # type: ignore[valid-type]


class CustomWellAttrs(WellAttrs04):
    well: CustomWellMeta  # type: ignore[valid-type]



from pathlib import Path

import pytest
import zarr

from ngio import create_empty_well, open_ome_zarr_well
from ngio.utils import NgioValueError


def test_open_real_ome_zarr_well(cardiomyocyte_tiny_path: Path):
    cardiomyocyte_tiny_path = cardiomyocyte_tiny_path
    cardiomyocyte_tiny_path = cardiomyocyte_tiny_path / "B" / "03"
    ome_zarr_well = open_ome_zarr_well(cardiomyocyte_tiny_path)
    assert isinstance(ome_zarr_well.__repr__(), str)
    assert ome_zarr_well.paths() == ["0"]
    assert ome_zarr_well.acquisition_ids == []


def test_create_and_edit_well(tmp_path: Path):
    test_well = create_empty_well(tmp_path / "test_well.zarr")
    assert test_well.paths() == []
    assert test_well.acquisition_ids == []

    test_well.add_image(image_path="0", acquisition_id=0, strict=False)
    test_well.add_image(image_path="1", acquisition_id=0, strict=True)

    with pytest.raises(NgioValueError):
        test_well.add_image(image_path="1", acquisition_id=1)

    test_well.atomic_add_image(image_path="2", acquisition_id=1, strict=False)
    assert len(test_well.paths()) == 3
    assert test_well.acquisition_ids == [0, 1], test_well.acquisition_ids

    store = test_well.get_image_store("0")
    assert isinstance(store, zarr.Group)



from collections.abc import Collection, Iterable
from typing import Literal

import dask
import dask.delayed
import numpy as np
import zarr

from ngio.common._axes_transforms import transform_dask_array, transform_numpy_array
from ngio.common._common_types import ArrayLike
from ngio.common._dimensions import Dimensions
from ngio.common._slicer import (
    SliceTransform,
    compute_and_slices,
    dask_get_slice,
    dask_set_slice,
    numpy_get_slice,
    numpy_set_slice,
)
from ngio.ome_zarr_meta.ngio_specs import AxesTransformation
from ngio.utils import NgioValueError


def _compute_from_disk_transforms(
    *,
    dimensions: Dimensions,
    axes_order: Collection[str] | None = None,
    **slice_kwargs: slice | int | Iterable[int],
) -> tuple[SliceTransform, tuple[AxesTransformation, ...]]:
    slices = compute_and_slices(dimensions=dimensions, **slice_kwargs)

    if axes_order is None:
        return slices, ()

    additional_transformations = dimensions._axes_mapper.to_order(axes_order)
    return slices, additional_transformations


def _compute_to_disk_transforms(
    *,
    dimensions: Dimensions,
    axes_order: Collection[str] | None = None,
    **slice_kwargs: slice | int | Iterable[int],
) -> tuple[SliceTransform, tuple[AxesTransformation, ...]]:
    slices = compute_and_slices(dimensions=dimensions, **slice_kwargs)
    if axes_order is None:
        return slices, ()

    additional_transformations = dimensions._axes_mapper.from_order(axes_order)
    return slices, additional_transformations


def _numpy_get_pipe(
    array: zarr.Array,
    slices: SliceTransform,
    transformations: tuple[AxesTransformation, ...],
) -> np.ndarray:
    array = numpy_get_slice(array, slices)
    return transform_numpy_array(array, transformations)


def _delayed_numpy_get_pipe(
    array: zarr.Array,
    slices: SliceTransform,
    transformations: tuple[AxesTransformation, ...],
) -> dask.delayed:
    array = dask.delayed(numpy_get_slice)(array, slices)
    return dask.delayed(transform_numpy_array)(array, transformations)


def _dask_get_pipe(
    array: zarr.Array,
    slices: SliceTransform,
    transformations: tuple[AxesTransformation, ...],
) -> dask.array:
    array = dask_get_slice(array, slices)
    return transform_dask_array(array, transformations)


def _numpy_set_pipe(
    array: zarr.Array,
    patch: np.ndarray,
    slices: SliceTransform,
    transformations: tuple[AxesTransformation, ...],
) -> None:
    patch = transform_numpy_array(patch, transformations)
    numpy_set_slice(array, patch, slices)


def _dask_set_pipe(
    array: zarr.Array,
    patch: np.ndarray,
    slices: SliceTransform,
    transformations: tuple[AxesTransformation, ...],
) -> None:
    patch = transform_dask_array(patch, transformations)
    dask_set_slice(array, patch, slices)


def _delayed_numpy_set_pipe(
    array: zarr.Array,
    patch: np.ndarray,
    slices: SliceTransform,
    transformations: tuple[AxesTransformation, ...],
) -> dask.delayed:
    patch = dask.delayed(transform_numpy_array)(patch, transformations)
    return dask.delayed(numpy_set_slice)(array, patch, slices)


def get_pipe(
    array: zarr.Array,
    *,
    dimensions: Dimensions,
    axes_order: Collection[str] | None = None,
    mode: Literal["numpy", "dask", "delayed"] = "numpy",
    **slice_kwargs: slice | int | Iterable[int],
):
    slices, transformations = _compute_from_disk_transforms(
        dimensions=dimensions, axes_order=axes_order, **slice_kwargs
    )
    match mode:
        case "numpy":
            return _numpy_get_pipe(array, slices, transformations)
        case "dask":
            return _dask_get_pipe(array, slices, transformations)

        case "delayed":
            return _delayed_numpy_get_pipe(array, slices, transformations)

        case _:
            raise NgioValueError(
                f"Unknown get pipe mode {mode}, expected 'numpy', 'dask' or 'delayed'."
            )


def set_pipe(
    array: zarr.Array,
    patch: ArrayLike,
    *,
    dimensions: Dimensions,
    axes_order: Collection[str] | None = None,
    **slice_kwargs: slice | int | Iterable[int],
):
    slices, transformations = _compute_to_disk_transforms(
        dimensions=dimensions, axes_order=axes_order, **slice_kwargs
    )
    if isinstance(patch, dask.array.Array):
        _dask_set_pipe(
            array=array, patch=patch, slices=slices, transformations=transformations
        )
    elif isinstance(patch, np.ndarray):
        _numpy_set_pipe(
            array=array, patch=patch, slices=slices, transformations=transformations
        )
    elif isinstance(patch, dask.delayed.Delayed):
        _delayed_numpy_set_pipe(
            array=array, patch=patch, slices=slices, transformations=transformations
        )
    else:
        raise NgioValueError("Unknown patch type, expected numpy, dask or delayed.")


def _mask_pipe_common(
    array: zarr.Array,
    label_array: zarr.Array,
    label: int,
    *,
    dimensions_array: Dimensions,
    dimensions_label: Dimensions,
    axes_order: Collection[str] | None = None,
    mode: Literal["numpy", "dask", "delayed"] = "numpy",
    **slice_kwargs: slice | int | Iterable[int],
):
    array_patch = get_pipe(
        array,
        dimensions=dimensions_array,
        axes_order=axes_order,
        mode=mode,
        **slice_kwargs,
    )

    if not dimensions_label.has_axis("c"):
        # Remove the 'c' from the slice_kwargs
        # This will not work if the query uses non-default
        # axes names for channel
        slice_kwargs = {k: v for k, v in slice_kwargs.items() if k != "c"}

    label_patch = get_pipe(
        label_array,
        dimensions=dimensions_label,
        axes_order=axes_order,
        mode=mode,
        **slice_kwargs,
    )

    if isinstance(array_patch, np.ndarray):
        label_patch = np.broadcast_to(label_patch, array_patch.shape)
    elif isinstance(array_patch, dask.array.Array):
        label_patch = dask.array.broadcast_to(label_patch, array_patch.shape)
    else:
        raise NgioValueError(f"Mode {mode} not yet supported for masked array.")

    mask = label_patch == label
    return array_patch, mask


def get_masked_pipe(
    array: zarr.Array,
    label_array: zarr.Array,
    label: int,
    *,
    dimensions_array: Dimensions,
    dimensions_label: Dimensions,
    axes_order: Collection[str] | None = None,
    mode: Literal["numpy", "dask", "delayed"] = "numpy",
    **slice_kwargs: slice | int | Iterable[int],
):
    array_patch, mask = _mask_pipe_common(
        array=array,
        label_array=label_array,
        label=label,
        dimensions_array=dimensions_array,
        dimensions_label=dimensions_label,
        axes_order=axes_order,
        mode=mode,
        **slice_kwargs,
    )
    array_patch[~mask] = 0
    return array_patch


def set_masked_pipe(
    array: zarr.Array,
    label_array: zarr.Array,
    label: int,
    patch: ArrayLike,
    *,
    dimensions_array: Dimensions,
    dimensions_label: Dimensions,
    axes_order: Collection[str] | None = None,
    **slice_kwargs: slice | int | Iterable[int],
):
    if isinstance(patch, dask.array.Array):
        mode = "dask"
    elif isinstance(patch, np.ndarray):
        mode = "numpy"
    else:
        raise NgioValueError(
            "Mode not yet supported for masked array. Expected a numpy or dask array."
        )

    array_patch, mask = _mask_pipe_common(
        array=array,
        label_array=label_array,
        label=label,
        dimensions_array=dimensions_array,
        dimensions_label=dimensions_label,
        axes_order=axes_order,
        mode=mode,
        **slice_kwargs,
    )
    patch = np.where(mask, patch, array_patch)
    set_pipe(
        array, patch, dimensions=dimensions_array, axes_order=axes_order, **slice_kwargs
    )



"""Base class for handling OME-NGFF metadata in Zarr groups."""

from typing import Generic, Protocol, TypeVar

from pydantic import ValidationError

from ngio.ome_zarr_meta.ngio_specs import (
    AxesSetup,
    NgioImageMeta,
    NgioLabelMeta,
    NgioPlateMeta,
    NgioWellMeta,
)
from ngio.ome_zarr_meta.v04 import (
    ngio_to_v04_image_meta,
    ngio_to_v04_label_meta,
    ngio_to_v04_plate_meta,
    ngio_to_v04_well_meta,
    v04_to_ngio_image_meta,
    v04_to_ngio_label_meta,
    v04_to_ngio_plate_meta,
    v04_to_ngio_well_meta,
)
from ngio.utils import (
    NgioValidationError,
    NgioValueError,
    ZarrGroupHandler,
)

ConverterError = ValidationError | Exception | None

###########################################################################
#
# The code below implements a generic class for handling OME-Zarr metadata
# in Zarr groups.
#
###########################################################################


class ImageMetaImporter(Protocol):
    @staticmethod
    def __call__(
        metadata: dict,
        axes_setup: AxesSetup | None = None,
        allow_non_canonical_axes: bool = False,
        strict_canonical_order: bool = True,
    ) -> tuple[bool, NgioImageMeta | ConverterError]:
        """Convert the metadata to a NgioImageMeta object.

        Args:
            metadata (dict): The metadata (typically from a Zarr group .attrs).
            axes_setup (AxesSetup, optional): The axes setup.
                This is used to map axes with non-canonical names.
            allow_non_canonical_axes (bool, optional): Whether to allow non-canonical
                axes.
            strict_canonical_order (bool, optional): Whether to enforce a strict
                canonical order.

        Returns:
            tuple[bool, NgioImageMeta | ConverterError]: A tuple with a boolean
                indicating whether the conversion was successful and the
                NgioImageMeta object or an error.

        """
        ...


class ImageMetaExporter(Protocol):
    def __call__(self, metadata: NgioImageMeta) -> dict: ...


class LabelMetaImporter(Protocol):
    @staticmethod
    def __call__(
        metadata: dict,
        axes_setup: AxesSetup | None = None,
        allow_non_canonical_axes: bool = False,
        strict_canonical_order: bool = True,
    ) -> tuple[bool, NgioLabelMeta | ConverterError]:
        """Convert the metadata to a NgioLabelMeta object.

        Args:
            metadata (dict): The metadata (typically from a Zarr group .attrs).
            axes_setup (AxesSetup, optional): The axes setup.
                This is used to map axes with non-canonical names.
            allow_non_canonical_axes (bool, optional): Whether to allow non-canonical
                axes.
            strict_canonical_order (bool, optional): Whether to enforce a strict
                canonical order.

        Returns:
            tuple[bool, NgioLabelMeta | ConverterError]: A tuple with a boolean
                indicating whether the conversion was successful and the
                NgioLabelMeta object or an error.

        """
        ...


class LabelMetaExporter(Protocol):
    def __call__(self, metadata: NgioLabelMeta) -> dict: ...


class WellMetaImporter(Protocol):
    def __call__(
        self, metadata: dict
    ) -> tuple[bool, NgioWellMeta | ConverterError]: ...


class WellMetaExporter(Protocol):
    def __call__(self, metadata: NgioWellMeta) -> dict: ...


class PlateMetaImporter(Protocol):
    def __call__(
        self, metadata: dict
    ) -> tuple[bool, NgioPlateMeta | ConverterError]: ...


class PlateMetaExporter(Protocol):
    def __call__(self, metadata: NgioPlateMeta) -> dict: ...


###########################################################################
#
# Image and label metadata handlers
#
###########################################################################

_image_meta = TypeVar("_image_meta", NgioImageMeta, NgioLabelMeta)
_image_meta_importer = TypeVar(
    "_image_meta_importer", ImageMetaImporter, LabelMetaImporter
)
_image_meta_exporter = TypeVar(
    "_image_meta_exporter", ImageMetaExporter, LabelMetaExporter
)


class GenericMetaHandler(
    Generic[_image_meta, _image_meta_importer, _image_meta_exporter]
):
    """Generic class for handling OME-Zarr metadata in Zarr groups."""

    def __init__(
        self,
        meta_importer: _image_meta_importer,
        meta_exporter: _image_meta_exporter,
        group_handler: ZarrGroupHandler,
        axes_setup: AxesSetup | None = None,
        allow_non_canonical_axes: bool = False,
        strict_canonical_order: bool = True,
    ):
        """Initialize the handler.

        Args:
            meta_importer (MetaImporter): The metadata importer.
            meta_exporter (MetaExporter): The metadata exporter.
            group_handler (ZarrGroupHandler): The Zarr group handler.
            axes_setup (AxesSetup, optional): The axes setup.
                This is used to map axes with non-canonical names.
            allow_non_canonical_axes (bool, optional): Whether to allow non-canonical
                axes.
            strict_canonical_order (bool, optional): Whether to enforce a strict
                canonical order.
        """
        self._group_handler = group_handler
        self._meta_importer = meta_importer
        self._meta_exporter = meta_exporter
        self._axes_setup = axes_setup
        self._allow_non_canonical_axes = allow_non_canonical_axes
        self._strict_canonical_order = strict_canonical_order

    def _load_meta(self, return_error: bool = False):
        """Load the metadata from the store."""
        attrs = self._group_handler.load_attrs()
        is_valid, meta_or_error = self._meta_importer(
            metadata=attrs,
            axes_setup=self._axes_setup,
            allow_non_canonical_axes=self._allow_non_canonical_axes,
            strict_canonical_order=self._strict_canonical_order,
        )
        if is_valid:
            return meta_or_error

        if return_error:
            return meta_or_error

        raise NgioValueError(f"Could not load metadata: {meta_or_error}")

    def safe_load_meta(self) -> _image_meta | ConverterError:
        """Load the metadata from the store."""
        return self._load_meta(return_error=True)

    def _write_meta(self, meta) -> None:
        """Write the metadata to the store."""
        _meta = self._meta_exporter(metadata=meta)
        self._group_handler.write_attrs(_meta)

    def write_meta(self, meta: _image_meta) -> None:
        self._write_meta(meta)

    @property
    def meta(self) -> _image_meta:
        """Return the metadata."""
        raise NotImplementedError("This method should be implemented in a subclass.")


class ImageMetaHandler(
    GenericMetaHandler[NgioImageMeta, ImageMetaImporter, ImageMetaExporter]
):
    """Generic class for handling OME-Zarr metadata in Zarr groups."""

    @property
    def meta(self) -> NgioImageMeta:
        meta = self._load_meta()
        if isinstance(meta, NgioImageMeta):
            return meta
        raise NgioValueError(f"Could not load metadata: {meta}")


class LabelMetaHandler(
    GenericMetaHandler[NgioLabelMeta, LabelMetaImporter, LabelMetaExporter]
):
    """Generic class for handling OME-Zarr metadata in Zarr groups."""

    @property
    def meta(self) -> NgioLabelMeta:
        meta = self._load_meta()
        if isinstance(meta, NgioLabelMeta):
            return meta
        raise NgioValueError(f"Could not load metadata: {meta}")


###########################################################################
#
# Well and plate metadata handlers
#
###########################################################################

_hcs_meta = TypeVar("_hcs_meta", NgioWellMeta, NgioPlateMeta)
_hcs_meta_importer = TypeVar("_hcs_meta_importer", WellMetaImporter, PlateMetaImporter)
_hcs_meta_exporter = TypeVar("_hcs_meta_exporter", WellMetaExporter, PlateMetaExporter)


class GenericHCSMetaHandler(Generic[_hcs_meta, _hcs_meta_importer, _hcs_meta_exporter]):
    """Generic class for handling OME-Zarr metadata in Zarr groups."""

    def __init__(
        self,
        meta_importer: _hcs_meta_importer,
        meta_exporter: _hcs_meta_exporter,
        group_handler: ZarrGroupHandler,
    ):
        self._group_handler = group_handler
        self._meta_importer = meta_importer
        self._meta_exporter = meta_exporter

    def _load_meta(self, return_error: bool = False):
        """Load the metadata from the store."""
        attrs = self._group_handler.load_attrs()
        is_valid, meta_or_error = self._meta_importer(metadata=attrs)
        if is_valid:
            return meta_or_error

        if return_error:
            return meta_or_error

        raise NgioValueError(f"Could not load metadata: {meta_or_error}")

    def safe_load_meta(self) -> _hcs_meta | ConverterError:
        """Load the metadata from the store."""
        return self._load_meta(return_error=True)

    def _write_meta(self, meta) -> None:
        _meta = self._meta_exporter(metadata=meta)
        self._group_handler.write_attrs(_meta)

    def write_meta(self, meta: _hcs_meta) -> None:
        self._write_meta(meta)

    @property
    def meta(self) -> _hcs_meta:
        raise NotImplementedError("This method should be implemented in a subclass.")


class WellMetaHandler(
    GenericHCSMetaHandler[NgioWellMeta, WellMetaImporter, WellMetaExporter]
):
    """Generic class for handling OME-Zarr metadata in Zarr groups."""

    @property
    def meta(self) -> NgioWellMeta:
        meta = self._load_meta()
        if isinstance(meta, NgioWellMeta):
            return meta
        raise NgioValueError(f"Could not load metadata: {meta}")


class PlateMetaHandler(
    GenericHCSMetaHandler[NgioPlateMeta, PlateMetaImporter, PlateMetaExporter]
):
    """Generic class for handling OME-Zarr metadata in Zarr groups."""

    @property
    def meta(self) -> NgioPlateMeta:
        meta = self._load_meta()
        if isinstance(meta, NgioPlateMeta):
            return meta
        raise NgioValueError(f"Could not load metadata: {meta}")


###########################################################################
#
# Metadata importer/exporter registration & builder classes
#
###########################################################################


_meta_exporter = TypeVar(
    "_meta_exporter",
    ImageMetaExporter,
    LabelMetaExporter,
    WellMetaExporter,
    PlateMetaExporter,
)
_meta_importer = TypeVar(
    "_meta_importer",
    ImageMetaImporter,
    LabelMetaImporter,
    WellMetaImporter,
    PlateMetaImporter,
)


class _ImporterExporter(Generic[_meta_importer, _meta_exporter]):
    def __init__(
        self,
        version: str,
        importer: _meta_importer,
        exporter: _meta_exporter,
    ):
        self.importer = importer
        self.exporter = exporter
        self.version = version


ImageImporterExporter = _ImporterExporter[ImageMetaImporter, ImageMetaExporter]
LabelImporterExporter = _ImporterExporter[LabelMetaImporter, LabelMetaExporter]
WellImporterExporter = _ImporterExporter[WellMetaImporter, WellMetaExporter]
PlateImporterExporter = _ImporterExporter[PlateMetaImporter, PlateMetaExporter]

_importer_exporter = TypeVar(
    "_importer_exporter",
    ImageImporterExporter,
    LabelImporterExporter,
    WellImporterExporter,
    PlateImporterExporter,
)
_image_handler = TypeVar("_image_handler", ImageMetaHandler, LabelMetaHandler)
_hcs_handler = TypeVar("_hcs_handler", WellMetaHandler, PlateMetaHandler)


class ImplementedMetaImporterExporter:
    _instance = None
    _image_ie: dict[str, ImageImporterExporter]
    _label_ie: dict[str, LabelImporterExporter]
    _well_ie: dict[str, WellImporterExporter]
    _plate_ie: dict[str, PlateImporterExporter]

    def __new__(cls):
        """Create a new instance of the class if it does not exist."""
        if cls._instance is None:
            cls._instance = super().__new__(cls)
            cls._image_ie = {}
            cls._label_ie = {}
            cls._well_ie = {}
            cls._plate_ie = {}
        return cls._instance

    def _find_image_handler(
        self,
        group_handler: ZarrGroupHandler,
        axes_setup: AxesSetup | None = None,
        allow_non_canonical_axes: bool = False,
        strict_canonical_order: bool = True,
        _ie_name: str = "_image_ie",
        _handler: type[_image_handler] = ImageMetaHandler,
    ) -> _image_handler:
        """Get an image metadata handler."""
        _errors = {}

        dict_ie = self.__getattribute__(_ie_name)

        for ie in reversed(dict_ie.values()):
            handler = _handler(
                meta_importer=ie.importer,
                meta_exporter=ie.exporter,
                group_handler=group_handler,
                axes_setup=axes_setup,
                allow_non_canonical_axes=allow_non_canonical_axes,
                strict_canonical_order=strict_canonical_order,
            )
            meta = handler.safe_load_meta()
            if isinstance(meta, ValidationError):
                _errors[ie.version] = meta
                continue
            return handler

        raise NgioValidationError(
            f"Could not load metadata from any known version. Errors: {_errors}"
        )

    def find_image_handler(
        self,
        group_handler: ZarrGroupHandler,
        axes_setup: AxesSetup | None = None,
        allow_non_canonical_axes: bool = False,
        strict_canonical_order: bool = True,
    ) -> ImageMetaHandler:
        """Get an image metadata handler."""
        return self._find_image_handler(
            group_handler=group_handler,
            axes_setup=axes_setup,
            allow_non_canonical_axes=allow_non_canonical_axes,
            strict_canonical_order=strict_canonical_order,
            _ie_name="_image_ie",
            _handler=ImageMetaHandler,
        )

    def get_image_meta_handler(
        self,
        group_handler: ZarrGroupHandler,
        version: str,
        axes_setup: AxesSetup | None = None,
        allow_non_canonical_axes: bool = False,
        strict_canonical_order: bool = True,
    ) -> ImageMetaHandler:
        """Get an image metadata handler."""
        if version not in self._image_ie:
            raise NgioValueError(f"Image handler for version {version} does not exist.")

        image_ie = self._image_ie[version]
        return ImageMetaHandler(
            meta_importer=image_ie.importer,
            meta_exporter=image_ie.exporter,
            group_handler=group_handler,
            axes_setup=axes_setup,
            allow_non_canonical_axes=allow_non_canonical_axes,
            strict_canonical_order=strict_canonical_order,
        )

    def _register(
        self,
        version: str,
        importer: _importer_exporter,
        overwrite: bool = False,
        _ie_name: str = "_image_ie",
    ):
        """Register an importer/exporter."""
        ie_dict = self.__getattribute__(_ie_name)
        if version in ie_dict and not overwrite:
            raise NgioValueError(
                f"Importer/exporter for version {version} already exists. "
                "Use 'overwrite=True' to overwrite."
            )
        ie_dict[version] = importer

    def register_image_ie(
        self,
        version: str,
        importer: ImageMetaImporter,
        exporter: ImageMetaExporter,
        overwrite: bool = False,
    ):
        """Register an importer/exporter."""
        importer_exporter = ImageImporterExporter(
            version=version, importer=importer, exporter=exporter
        )
        self._register(
            version=version,
            importer=importer_exporter,
            overwrite=overwrite,
            _ie_name="_image_ie",
        )

    def find_label_handler(
        self,
        group_handler: ZarrGroupHandler,
        axes_setup: AxesSetup | None = None,
        allow_non_canonical_axes: bool = False,
        strict_canonical_order: bool = True,
    ) -> LabelMetaHandler:
        """Get a label metadata handler."""
        return self._find_image_handler(
            group_handler=group_handler,
            axes_setup=axes_setup,
            allow_non_canonical_axes=allow_non_canonical_axes,
            strict_canonical_order=strict_canonical_order,
            _ie_name="_label_ie",
            _handler=LabelMetaHandler,
        )

    def get_label_meta_handler(
        self,
        group_handler: ZarrGroupHandler,
        version: str,
        axes_setup: AxesSetup | None = None,
        allow_non_canonical_axes: bool = False,
        strict_canonical_order: bool = True,
    ) -> LabelMetaHandler:
        """Get a label metadata handler."""
        if version not in self._label_ie:
            raise NgioValueError(f"Label handler for version {version} does not exist.")

        label_ie = self._label_ie[version]
        return LabelMetaHandler(
            meta_importer=label_ie.importer,
            meta_exporter=label_ie.exporter,
            group_handler=group_handler,
            axes_setup=axes_setup,
            allow_non_canonical_axes=allow_non_canonical_axes,
            strict_canonical_order=strict_canonical_order,
        )

    def register_label_ie(
        self,
        version: str,
        importer: LabelMetaImporter,
        exporter: LabelMetaExporter,
        overwrite: bool = False,
    ):
        """Register an importer/exporter."""
        importer_exporter = LabelImporterExporter(
            version=version, importer=importer, exporter=exporter
        )
        self._register(
            version=version,
            importer=importer_exporter,
            overwrite=overwrite,
            _ie_name="_label_ie",
        )

    def _find_hcs_handler(
        self,
        group_handler: ZarrGroupHandler,
        _ie_name: str = "_well_ie",
        _handler: type[_hcs_handler] = WellMetaHandler,
    ) -> _hcs_handler:
        """Get a handler for a HCS metadata."""
        _errors = {}

        dict_ie = self.__getattribute__(_ie_name)

        for ie in reversed(dict_ie.values()):
            handler = _handler(
                meta_importer=ie.importer,
                meta_exporter=ie.exporter,
                group_handler=group_handler,
            )
            meta = handler.safe_load_meta()
            if isinstance(meta, ValidationError):
                _errors[ie.version] = meta
                continue
            return handler

        raise NgioValidationError(
            f"Could not load metadata from any known version. Errors: {_errors}"
        )

    def find_well_handler(
        self,
        group_handler: ZarrGroupHandler,
    ) -> WellMetaHandler:
        """Get a well metadata handler."""
        return self._find_hcs_handler(
            group_handler=group_handler,
            _ie_name="_well_ie",
            _handler=WellMetaHandler,
        )

    def get_well_meta_handler(
        self,
        group_handler: ZarrGroupHandler,
        version: str,
    ) -> WellMetaHandler:
        """Get a well metadata handler."""
        if version not in self._well_ie:
            raise NgioValueError(f"Well handler for version {version} does not exist.")

        well_ie = self._well_ie[version]
        return WellMetaHandler(
            meta_importer=well_ie.importer,
            meta_exporter=well_ie.exporter,
            group_handler=group_handler,
        )

    def register_well_ie(
        self,
        version: str,
        importer: WellMetaImporter,
        exporter: WellMetaExporter,
        overwrite: bool = False,
    ):
        """Register an importer/exporter."""
        importer_exporter = WellImporterExporter(
            version=version, importer=importer, exporter=exporter
        )
        self._register(
            version=version,
            importer=importer_exporter,
            overwrite=overwrite,
            _ie_name="_well_ie",
        )

    def find_plate_handler(
        self,
        group_handler: ZarrGroupHandler,
    ) -> PlateMetaHandler:
        """Get a plate metadata handler."""
        return self._find_hcs_handler(
            group_handler=group_handler,
            _ie_name="_plate_ie",
            _handler=PlateMetaHandler,
        )

    def get_plate_meta_handler(
        self,
        group_handler: ZarrGroupHandler,
        version: str,
    ) -> PlateMetaHandler:
        """Get a plate metadata handler."""
        if version not in self._plate_ie:
            raise NgioValueError(f"Plate handler for version {version} does not exist.")

        plate_ie = self._plate_ie[version]
        return PlateMetaHandler(
            meta_importer=plate_ie.importer,
            meta_exporter=plate_ie.exporter,
            group_handler=group_handler,
        )

    def register_plate_ie(
        self,
        version: str,
        importer: PlateMetaImporter,
        exporter: PlateMetaExporter,
        overwrite: bool = False,
    ):
        """Register an importer/exporter."""
        importer_exporter = PlateImporterExporter(
            version=version, importer=importer, exporter=exporter
        )
        self._register(
            version=version,
            importer=importer_exporter,
            overwrite=overwrite,
            _ie_name="_plate_ie",
        )


###########################################################################
#
# Register metadata importers/exporters
#
###########################################################################


ImplementedMetaImporterExporter().register_image_ie(
    version="0.4",
    importer=v04_to_ngio_image_meta,
    exporter=ngio_to_v04_image_meta,
)
ImplementedMetaImporterExporter().register_label_ie(
    version="0.4",
    importer=v04_to_ngio_label_meta,
    exporter=ngio_to_v04_label_meta,
)
ImplementedMetaImporterExporter().register_well_ie(
    version="0.4", importer=v04_to_ngio_well_meta, exporter=ngio_to_v04_well_meta
)
ImplementedMetaImporterExporter().register_plate_ie(
    version="0.4", importer=v04_to_ngio_plate_meta, exporter=ngio_to_v04_plate_meta
)


###########################################################################
#
# Public functions to avoid direct access to the importer/exporter
# registration methods
#
###########################################################################


def find_image_meta_handler(
    group_handler: ZarrGroupHandler,
    axes_setup: AxesSetup | None = None,
    allow_non_canonical_axes: bool = False,
    strict_canonical_order: bool = True,
) -> ImageMetaHandler:
    """Open an image metadata handler."""
    return ImplementedMetaImporterExporter().find_image_handler(
        group_handler=group_handler,
        axes_setup=axes_setup,
        allow_non_canonical_axes=allow_non_canonical_axes,
        strict_canonical_order=strict_canonical_order,
    )


def get_image_meta_handler(
    group_handler: ZarrGroupHandler,
    version: str,
    axes_setup: AxesSetup | None = None,
    allow_non_canonical_axes: bool = False,
    strict_canonical_order: bool = True,
) -> ImageMetaHandler:
    """Open an image metadata handler."""
    return ImplementedMetaImporterExporter().get_image_meta_handler(
        group_handler=group_handler,
        version=version,
        axes_setup=axes_setup,
        allow_non_canonical_axes=allow_non_canonical_axes,
        strict_canonical_order=strict_canonical_order,
    )


def find_label_meta_handler(
    group_handler: ZarrGroupHandler,
    axes_setup: AxesSetup | None = None,
    allow_non_canonical_axes: bool = False,
    strict_canonical_order: bool = True,
) -> LabelMetaHandler:
    """Open a label metadata handler."""
    return ImplementedMetaImporterExporter().find_label_handler(
        group_handler=group_handler,
        axes_setup=axes_setup,
        allow_non_canonical_axes=allow_non_canonical_axes,
        strict_canonical_order=strict_canonical_order,
    )


def get_label_meta_handler(
    group_handler: ZarrGroupHandler,
    version: str,
    axes_setup: AxesSetup | None = None,
    allow_non_canonical_axes: bool = False,
    strict_canonical_order: bool = True,
) -> LabelMetaHandler:
    """Open a label metadata handler."""
    return ImplementedMetaImporterExporter().get_label_meta_handler(
        group_handler=group_handler,
        version=version,
        axes_setup=axes_setup,
        allow_non_canonical_axes=allow_non_canonical_axes,
        strict_canonical_order=strict_canonical_order,
    )


def find_well_meta_handler(group_handler: ZarrGroupHandler) -> WellMetaHandler:
    """Open a well metadata handler."""
    return ImplementedMetaImporterExporter().find_well_handler(
        group_handler=group_handler,
    )


def get_well_meta_handler(
    group_handler: ZarrGroupHandler,
    version: str,
) -> WellMetaHandler:
    """Open a well metadata handler."""
    return ImplementedMetaImporterExporter().get_well_meta_handler(
        group_handler=group_handler,
        version=version,
    )


def find_plate_meta_handler(group_handler: ZarrGroupHandler) -> PlateMetaHandler:
    """Open a plate metadata handler."""
    return ImplementedMetaImporterExporter().find_plate_handler(
        group_handler=group_handler
    )


def get_plate_meta_handler(
    group_handler: ZarrGroupHandler,
    version: str,
) -> PlateMetaHandler:
    """Open a plate metadata handler."""
    return ImplementedMetaImporterExporter().get_plate_meta_handler(
        group_handler=group_handler,
        version=version,
    )



import fsspec.implementations.http


def fractal_fsspec_store(
    url: str, fractal_token: str | None = None, client_kwargs: dict | None = None
) -> fsspec.mapping.FSMap:
    """Simple function to get an http fsspec store from a url."""
    client_kwargs = {} if client_kwargs is None else client_kwargs
    if fractal_token is not None:
        client_kwargs["headers"] = {"Authorization": f"Bearer {fractal_token}"}
    fs = fsspec.implementations.http.HTTPFileSystem(client_kwargs=client_kwargs)
    store = fs.get_mapper(url)
    return store



"""Utilities to build masking regions of interest (ROIs)."""

import itertools

import dask
import dask.array as da
import dask.delayed
import numpy as np
import scipy.ndimage as ndi

from ngio.common._roi import Roi, RoiPixels
from ngio.ome_zarr_meta import PixelSize
from ngio.utils import NgioValueError


def _compute_offsets(chunks):
    """Given a chunks tuple, compute cumulative offsets for each axis.

    Returns a list where each element is a list of offsets for that dimension.
    """
    offsets = []
    for dim_chunks in chunks:
        dim_offsets = [0]
        for size in dim_chunks:
            dim_offsets.append(dim_offsets[-1] + size)
        offsets.append(dim_offsets)
    return offsets


def _adjust_slices(slices, offset):
    """Adjust slices to global coordinates using the provided offset."""
    adjusted_slices = {}
    for label, s in slices.items():
        adjusted = tuple(
            slice(s_dim.start + off, s_dim.stop + off)
            for s_dim, off in zip(s, offset, strict=True)
        )
        adjusted_slices[label] = adjusted
    return adjusted_slices


@dask.delayed
def _process_chunk(chunk, offset):
    """Process a single chunk.

    run ndi.find_objects and adjust the slices
    to global coordinates using the provided offset.
    """
    local_slices = compute_slices(chunk)
    local_slices = _adjust_slices(local_slices, offset)
    return local_slices


def _merge_slices(
    slice1: tuple[slice, ...], slice2: tuple[slice, ...]
) -> tuple[slice, ...]:
    """Merge two slices."""
    merged = []
    for s1, s2 in zip(slice1, slice2, strict=True):
        start = min(s1.start, s2.start)
        stop = max(s1.stop, s2.stop)
        merged.append(slice(start, stop))
    return tuple(merged)


@dask.delayed
def _collect_slices(
    local_slices: list[dict[int, tuple[slice, ...]]],
) -> dict[int, tuple[slice]]:
    """Collect the slices from the delayed results."""
    global_slices = {}
    for result in local_slices:
        for label, s in result.items():
            if label in global_slices:
                global_slices[label] = _merge_slices(global_slices[label], s)
            else:
                global_slices[label] = s
    return global_slices


def compute_slices(segmentation: np.ndarray) -> dict[int, tuple[slice, ...]]:
    """Compute slices for each label in a segmentation.

    Args:
        segmentation (ndarray): The segmentation array.

    Returns:
        dict[int, tuple[slice]]: A dictionary with the label as key
            and the slice as value.
    """
    slices = ndi.find_objects(segmentation)
    slices_dict = {}
    for label, s in enumerate(slices, start=1):
        if s is None:
            continue
        else:
            slices_dict[label] = s
    return slices_dict


def lazy_compute_slices(segmentation: da.Array) -> dict[int, tuple[slice, ...]]:
    """Compute slices for each label in a segmentation."""
    global_offsets = _compute_offsets(segmentation.chunks)
    delayed_chunks = segmentation.to_delayed()

    grid_shape = tuple(len(c) for c in segmentation.chunks)

    grid_indices = list(itertools.product(*[range(n) for n in grid_shape]))
    delayed_results = []
    for idx, chunk in zip(grid_indices, np.ravel(delayed_chunks), strict=True):
        offset = tuple(global_offsets[dim][idx[dim]] for dim in range(len(idx)))
        delayed_result = _process_chunk(chunk, offset)
        delayed_results.append(delayed_result)

    return _collect_slices(delayed_results).compute()


def compute_masking_roi(
    segmentation: np.ndarray | da.Array, pixel_size: PixelSize
) -> list[Roi]:
    """Compute a ROIs for each label in a segmentation.

    This function expects a 2D or 3D segmentation array.
    And this function expects the axes order to be 'zyx' or 'yx'.
    Other axes orders are not supported.

    """
    if segmentation.ndim not in [2, 3]:
        raise NgioValueError("Only 2D and 3D segmentations are supported.")

    if isinstance(segmentation, da.Array):
        slices = lazy_compute_slices(segmentation)
    else:
        slices = compute_slices(segmentation)

    rois = []
    for label, slice_ in slices.items():
        if len(slice_) == 2:
            min_z, min_y, min_x = 0, slice_[0].start, slice_[1].start
            max_z, max_y, max_x = 1, slice_[0].stop, slice_[1].stop
        elif len(slice_) == 3:
            min_z, min_y, min_x = slice_[0].start, slice_[1].start, slice_[2].start
            max_z, max_y, max_x = slice_[0].stop, slice_[1].stop, slice_[2].stop
        else:
            raise ValueError("Invalid slice length.")
        roi = RoiPixels(
            name=str(label),
            x_length=max_x - min_x,
            y_length=max_y - min_y,
            z_length=max_z - min_z,
            x=min_x,
            y=min_y,
            z=min_z,
        )

        roi = roi.to_roi(pixel_size)
        rois.append(roi)
    return rois



"""Fractal internal module for dataset metadata handling."""

import math
from functools import total_ordering

import numpy as np

from ngio.ome_zarr_meta.ngio_specs import (
    DefaultSpaceUnit,
    DefaultTimeUnit,
    SpaceUnits,
    TimeUnits,
)

################################################################################################
#
# PixelSize model
# The PixelSize model is used to store the pixel size in 3D space.
# The model does not store scaling factors and units for other axes.
#
#################################################################################################


def _validate_type(value: float, name: str) -> float:
    """Check the type of the value."""
    if not isinstance(value, int | float):
        raise TypeError(f"{name} must be a number.")
    return float(value)


@total_ordering
class PixelSize:
    """PixelSize class to store the pixel size in 3D space."""

    def __init__(
        self,
        x: float,
        y: float,
        z: float,
        t: float = 1,
        space_unit: SpaceUnits | str | None = DefaultSpaceUnit,
        time_unit: TimeUnits | str | None = DefaultTimeUnit,
    ):
        """Initialize the pixel size."""
        self.x = _validate_type(x, "x")
        self.y = _validate_type(y, "y")
        self.z = _validate_type(z, "z")
        self.t = _validate_type(t, "t")

        self._space_unit = space_unit
        self._time_unit = time_unit

    def __repr__(self) -> str:
        """Return a string representation of the pixel size."""
        return f"PixelSize(x={self.x}, y={self.y}, z={self.z}, t={self.t})"

    def __eq__(self, other) -> bool:
        """Check if two pixel sizes are equal."""
        if not isinstance(other, PixelSize):
            raise TypeError("Can only compare PixelSize with PixelSize.")

        if (
            self.time_unit is not None
            and other.time_unit is None
            and self.time_unit != other.time_unit
        ):
            return False

        if self.space_unit != other.space_unit:
            return False
        return math.isclose(self.distance(other), 0)

    def __lt__(self, other: "PixelSize") -> bool:
        """Check if one pixel size is less than the other."""
        if not isinstance(other, PixelSize):
            raise TypeError("Can only compare PixelSize with PixelSize.")
        ref = PixelSize(
            0,
            0,
            0,
            0,
            space_unit=self.space_unit,
            time_unit=self.time_unit,  # type: ignore
        )
        return self.distance(ref) < other.distance(ref)

    def as_dict(self) -> dict:
        """Return the pixel size as a dictionary."""
        return {"t": self.t, "z": self.z, "y": self.y, "x": self.x}

    @property
    def space_unit(self) -> SpaceUnits | str | None:
        """Return the space unit."""
        return self._space_unit

    @property
    def time_unit(self) -> TimeUnits | str | None:
        """Return the time unit."""
        return self._time_unit

    @property
    def tzyx(self) -> tuple[float, float, float, float]:
        """Return the voxel size in t, z, y, x order."""
        return self.t, self.z, self.y, self.x

    @property
    def zyx(self) -> tuple[float, float, float]:
        """Return the voxel size in z, y, x order."""
        return self.z, self.y, self.x

    @property
    def yx(self) -> tuple[float, float]:
        """Return the xy plane pixel size in y, x order."""
        return self.y, self.x

    @property
    def voxel_volume(self) -> float:
        """Return the volume of a voxel."""
        return self.y * self.x * self.z

    @property
    def xy_plane_area(self) -> float:
        """Return the area of the xy plane."""
        return self.y * self.x

    @property
    def time_spacing(self) -> float | None:
        """Return the time spacing."""
        return self.t

    def distance(self, other: "PixelSize") -> float:
        """Return the distance between two pixel sizes."""
        return float(np.linalg.norm(np.array(self.tzyx) - np.array(other.tzyx)))



"""Implementation of the ROI Table class.

This class follows the roi_table specification at:
https://fractal-analytics-platform.github.io/fractal-tasks-core/tables/
"""

# Import _type to avoid name conflict with table.type
from builtins import type as _type
from collections.abc import Iterable
from typing import Generic, Literal, TypeVar

import pandas as pd
from pydantic import BaseModel

from ngio.common import Roi
from ngio.tables._validators import validate_columns
from ngio.tables.backends import BackendMeta, ImplementedTableBackends
from ngio.utils import NgioValueError, ZarrGroupHandler

REQUIRED_COLUMNS = [
    "x_micrometer",
    "y_micrometer",
    "z_micrometer",
    "len_x_micrometer",
    "len_y_micrometer",
    "len_z_micrometer",
]


ORIGIN_COLUMNS = [
    "x_micrometer_original",
    "y_micrometer_original",
]

TRANSLATION_COLUMNS = ["translation_x", "translation_y", "translation_z"]

OPTIONAL_COLUMNS = ORIGIN_COLUMNS + TRANSLATION_COLUMNS


def _dataframe_to_rois(dataframe: pd.DataFrame) -> dict[str, Roi]:
    """Convert a DataFrame to a WorldCooROI object."""
    rois = {}
    for key, row in dataframe.iterrows():
        # check if optional columns are present
        origin = {col: row.get(col, None) for col in ORIGIN_COLUMNS}
        origin = dict(filter(lambda x: x[1] is not None, origin.items()))
        translation = {col: row.get(col, None) for col in TRANSLATION_COLUMNS}
        translation = dict(filter(lambda x: x[1] is not None, translation.items()))

        roi = Roi(
            name=str(key),
            x=row["x_micrometer"],
            y=row["y_micrometer"],
            z=row["z_micrometer"],
            x_length=row["len_x_micrometer"],
            y_length=row["len_y_micrometer"],
            z_length=row["len_z_micrometer"],
            unit="micrometer",  # type: ignore
            **origin,
            **translation,
        )
        rois[roi.name] = roi
    return rois


def _rois_to_dataframe(rois: dict[str, Roi], index_key: str) -> pd.DataFrame:
    """Convert a list of WorldCooROI objects to a DataFrame."""
    data = []
    for roi in rois.values():
        row = {
            index_key: roi.name,
            "x_micrometer": roi.x,
            "y_micrometer": roi.y,
            "z_micrometer": roi.z,
            "len_x_micrometer": roi.x_length,
            "len_y_micrometer": roi.y_length,
            "len_z_micrometer": roi.z_length,
        }

        extra = roi.model_extra or {}
        for col in ORIGIN_COLUMNS:
            if col in extra:
                row[col] = extra[col]

        for col in TRANSLATION_COLUMNS:
            if col in extra:
                row[col] = extra[col]
        data.append(row)
    dataframe = pd.DataFrame(data)
    dataframe = dataframe.set_index(index_key)
    return dataframe


class RoiTableV1Meta(BackendMeta):
    """Metadata for the ROI table."""

    fractal_table_version: Literal["1"] = "1"
    type: Literal["roi_table"] = "roi_table"


class RegionMeta(BaseModel):
    """Metadata for the region."""

    path: str


class MaskingRoiTableV1Meta(BackendMeta):
    """Metadata for the ROI table."""

    fractal_table_version: Literal["1"] = "1"
    type: Literal["masking_roi_table"] = "masking_roi_table"
    region: RegionMeta | None = None
    instance_key: str = "label"


_roi_meta = TypeVar("_roi_meta", RoiTableV1Meta, MaskingRoiTableV1Meta)


class _GenericRoiTableV1(Generic[_roi_meta]):
    """Class to a non-specific table."""

    _meta: _roi_meta

    def __init__(
        self, meta: _roi_meta | None = None, rois: Iterable[Roi] | None = None
    ) -> None:
        """Create a new ROI table."""
        if meta is None:
            raise NgioValueError("Metadata must be provided.")
        self._meta = meta
        self._table_backend = None

        self._rois = {}
        if rois is not None:
            self.add(rois)

    @staticmethod
    def type() -> str:
        """Return the type of the table."""
        raise NotImplementedError

    @staticmethod
    def version() -> Literal["1"]:
        """Return the version of the fractal table."""
        return "1"

    @staticmethod
    def _index_key() -> str:
        """Return the index key of the table."""
        raise NotImplementedError

    @staticmethod
    def _index_type() -> Literal["int", "str"]:
        """Return the index type of the table."""
        raise NotImplementedError

    @staticmethod
    def _meta_type() -> _type[_roi_meta]:
        """Return the metadata type of the table."""
        raise NotImplementedError

    @property
    def backend_name(self) -> str | None:
        """Return the name of the backend."""
        if self._table_backend is None:
            return None
        return self._table_backend.backend_name()

    @classmethod
    def _from_handler(
        cls, handler: ZarrGroupHandler, backend_name: str | None = None
    ) -> "_GenericRoiTableV1":
        """Create a new ROI table from a Zarr store."""
        meta = cls._meta_type()(**handler.load_attrs())

        if backend_name is None:
            backend = ImplementedTableBackends().get_backend(
                backend_name=meta.backend,
                group_handler=handler,
                index_key=cls._index_key(),
                index_type=cls._index_type(),
            )
        else:
            backend = ImplementedTableBackends().get_backend(
                backend_name=backend_name,
                group_handler=handler,
                index_key=cls._index_key(),
                index_type=cls._index_type(),
            )
            meta.backend = backend_name

        if not backend.implements_pandas:
            raise NgioValueError(
                "The backend does not implement the dataframe protocol."
            )

        # This will be implemented in the child classes
        table = cls()
        table._meta = meta
        table._table_backend = backend

        dataframe = backend.load_as_pandas_df()
        dataframe = validate_columns(
            dataframe,
            required_columns=REQUIRED_COLUMNS,
            optional_columns=OPTIONAL_COLUMNS,
        )
        table._rois = _dataframe_to_rois(dataframe)
        return table

    def _set_backend(
        self,
        handler: ZarrGroupHandler,
        backend_name: str | None = None,
    ) -> None:
        """Set the backend of the table."""
        backend = ImplementedTableBackends().get_backend(
            backend_name=backend_name,
            group_handler=handler,
            index_key=self._index_key(),
            index_type=self._index_type(),
        )
        self._meta.backend = backend_name
        self._table_backend = backend

    def rois(self) -> list[Roi]:
        """List all ROIs in the table."""
        return list(self._rois.values())

    def add(self, roi: Roi | Iterable[Roi], overwrite: bool = False) -> None:
        """Append ROIs to the current table.

        Args:
            roi: A single ROI or a list of ROIs to add to the table.
            overwrite: If True, overwrite existing ROIs with the same name.
        """
        if isinstance(roi, Roi):
            roi = [roi]

        for _roi in roi:
            if not overwrite and _roi.name in self._rois:
                raise NgioValueError(f"ROI {_roi.name} already exists in the table.")
            self._rois[_roi.name] = _roi

    def consolidate(self) -> None:
        """Write the current state of the table to the Zarr file."""
        if self._table_backend is None:
            raise NgioValueError(
                "No backend set for the table. "
                "Please add the table to a OME-Zarr Image before calling consolidate."
            )

        dataframe = _rois_to_dataframe(self._rois, index_key=self._index_key())
        dataframe = validate_columns(
            dataframe,
            required_columns=REQUIRED_COLUMNS,
            optional_columns=OPTIONAL_COLUMNS,
        )
        self._table_backend.write(
            dataframe, metadata=self._meta.model_dump(exclude_none=True), mode="pandas"
        )


class RoiTableV1(_GenericRoiTableV1[RoiTableV1Meta]):
    """Class to handle fractal ROI tables.

    To know more about the ROI table format, please refer to the
    specification at:
    https://fractal-analytics-platform.github.io/fractal-tasks-core/tables/
    """

    def __init__(self, rois: Iterable[Roi] | None = None) -> None:
        """Create a new ROI table."""
        super().__init__(RoiTableV1Meta(), rois)

    def __repr__(self) -> str:
        """Return a string representation of the table."""
        prop = f"num_rois={len(self._rois)}"
        return f"RoiTableV1({prop})"

    @staticmethod
    def type() -> Literal["roi_table"]:
        """Return the type of the table."""
        return "roi_table"

    @staticmethod
    def _index_key() -> str:
        """Return the index key of the table."""
        return "FieldIndex"

    @staticmethod
    def _index_type() -> Literal["int", "str"]:
        """Return the index type of the table."""
        return "str"

    @staticmethod
    def _meta_type() -> _type[RoiTableV1Meta]:
        """Return the metadata type of the table."""
        return RoiTableV1Meta

    def get(self, roi_name: str) -> Roi:
        """Get an ROI from the table."""
        if roi_name not in self._rois:
            raise NgioValueError(f"ROI {roi_name} not found in the table.")
        return self._rois[roi_name]


class MaskingRoiTableV1(_GenericRoiTableV1[MaskingRoiTableV1Meta]):
    """Class to handle fractal ROI tables.

    To know more about the ROI table format, please refer to the
    specification at:
    https://fractal-analytics-platform.github.io/fractal-tasks-core/tables/
    """

    def __init__(
        self,
        rois: Iterable[Roi] | None = None,
        reference_label: str | None = None,
    ) -> None:
        """Create a new ROI table."""
        meta = MaskingRoiTableV1Meta()
        if reference_label is not None:
            meta.region = RegionMeta(path=reference_label)
        super().__init__(meta, rois)

    def __repr__(self) -> str:
        """Return a string representation of the table."""
        prop = f"num_rois={len(self._rois)}"
        if self.reference_label is not None:
            prop += f", reference_label={self.reference_label}"
        return f"MaskingRoiTableV1({prop})"

    @staticmethod
    def type() -> Literal["masking_roi_table"]:
        """Return the type of the table."""
        return "masking_roi_table"

    @staticmethod
    def _index_key() -> str:
        """Return the index key of the table."""
        return "label"

    @staticmethod
    def _index_type() -> Literal["int", "str"]:
        """Return the index type of the table."""
        return "int"

    @staticmethod
    def _meta_type() -> _type[MaskingRoiTableV1Meta]:
        """Return the metadata type of the table."""
        return MaskingRoiTableV1Meta

    @property
    def reference_label(self) -> str | None:
        """Return the reference label."""
        path = self._meta.region
        if path is None:
            return None
        path = path.path
        path = path.split("/")[-1]
        return path

    def get(self, label: int) -> Roi:
        """Get an ROI from the table."""
        _label = str(label)
        if _label not in self._rois:
            raise KeyError(f"ROI {_label} not found in the table.")
        return self._rois[_label]



"""Fractal internal module for dataset metadata handling."""

from collections.abc import Collection

from ngio.ome_zarr_meta.ngio_specs._axes import (
    AxesMapper,
    AxesSetup,
    Axis,
    AxisType,
    DefaultSpaceUnit,
    DefaultTimeUnit,
    SpaceUnits,
    TimeUnits,
)
from ngio.ome_zarr_meta.ngio_specs._pixel_size import PixelSize
from ngio.utils import NgioValidationError


class Dataset:
    """Model for a dataset in the multiscale.

    To initialize the Dataset object, the path, the axes, scale, and translation list
    can be provided with on_disk order.
    """

    def __init__(
        self,
        *,
        # args coming from ngff specs
        path: str,
        on_disk_axes: Collection[Axis],
        on_disk_scale: Collection[float],
        on_disk_translation: Collection[float] | None = None,
        # user defined args
        axes_setup: AxesSetup | None = None,
        allow_non_canonical_axes: bool = False,
        strict_canonical_order: bool = False,
    ):
        """Initialize the Dataset object.

        Args:
            path (str): The path of the dataset.
            on_disk_axes (list[Axis]): The list of axes in the multiscale.
            on_disk_scale (list[float]): The list of scale transformation.
                The scale transformation must have the same length as the axes.
            on_disk_translation (list[float] | None): The list of translation.
            axes_setup (AxesSetup): The axes setup object
            allow_non_canonical_axes (bool): Allow non-canonical axes.
            strict_canonical_order (bool): Strict canonical order.
        """
        self._path = path
        self._axes_mapper = AxesMapper(
            on_disk_axes=on_disk_axes,
            axes_setup=axes_setup,
            allow_non_canonical_axes=allow_non_canonical_axes,
            strict_canonical_order=strict_canonical_order,
        )

        if len(on_disk_scale) != len(on_disk_axes):
            raise NgioValidationError(
                "The length of the scale transformation must be the same as the axes."
            )
        self._on_disk_scale = list(on_disk_scale)

        on_disk_translation = on_disk_translation or [0.0] * len(on_disk_axes)
        if len(on_disk_translation) != len(on_disk_axes):
            raise NgioValidationError(
                "The length of the translation must be the same as the axes."
            )
        self._on_disk_translation = list(on_disk_translation)

    def get_scale(self, axis_name: str) -> float:
        """Return the scale for a given axis."""
        idx = self._axes_mapper.get_index(axis_name)
        if idx is None:
            return 1.0
        return self._on_disk_scale[idx]

    def get_translation(self, axis_name: str) -> float:
        """Return the translation for a given axis."""
        idx = self._axes_mapper.get_index(axis_name)
        if idx is None:
            return 0.0
        return self._on_disk_translation[idx]

    @property
    def path(self) -> str:
        """Return the path of the dataset."""
        return self._path

    @property
    def space_unit(self) -> str | None:
        """Return the space unit for a given axis."""
        x_axis = self._axes_mapper.get_axis("x")
        y_axis = self._axes_mapper.get_axis("y")

        if x_axis is None or y_axis is None:
            raise NgioValidationError(
                "The dataset must have x and y axes to determine the space unit."
            )

        if x_axis.unit == y_axis.unit:
            return x_axis.unit
        else:
            raise NgioValidationError(
                "Inconsistent space units. "
                f"x={x_axis.unit} and y={y_axis.unit} should have the same unit."
            )

    @property
    def time_unit(self) -> str | None:
        """Return the time unit for a given axis."""
        t_axis = self._axes_mapper.get_axis("t")
        if t_axis is None:
            return None
        return t_axis.unit

    @property
    def pixel_size(self) -> PixelSize:
        """Return the pixel size for the dataset."""
        return PixelSize(
            x=self.get_scale("x"),
            y=self.get_scale("y"),
            z=self.get_scale("z"),
            t=self.get_scale("t"),
            space_unit=self.space_unit,  # type: ignore
            time_unit=self.time_unit,  # type: ignore
        )

    @property
    def axes_mapper(self) -> AxesMapper:
        """Return the axes mapper object."""
        return self._axes_mapper

    def to_units(
        self,
        *,
        space_unit: SpaceUnits = DefaultSpaceUnit,
        time_unit: TimeUnits = DefaultTimeUnit,
    ) -> "Dataset":
        """Convert the pixel size to the given units.

        Args:
            space_unit(str): The space unit to convert to.
            time_unit(str): The time unit to convert to.
        """
        new_axes = []
        for ax in self.axes_mapper.on_disk_axes:
            if ax.axis_type == AxisType.space:
                new_ax = Axis(
                    on_disk_name=ax.on_disk_name,
                    axis_type=ax.axis_type,
                    unit=space_unit,
                )
                new_axes.append(new_ax)
            elif ax.axis_type == AxisType.time:
                new_ax = Axis(
                    on_disk_name=ax.on_disk_name, axis_type=ax.axis_type, unit=time_unit
                )
                new_axes.append(new_ax)
            else:
                new_axes.append(ax)

        return Dataset(
            path=self.path,
            on_disk_axes=new_axes,
            on_disk_scale=self._on_disk_scale,
            on_disk_translation=self._on_disk_translation,
            axes_setup=self.axes_mapper.axes_setup,
            allow_non_canonical_axes=self.axes_mapper.allow_non_canonical_axes,
            strict_canonical_order=self.axes_mapper.strict_canonical_order,
        )



"""Common utilities for working with Zarr groups in consistent ways."""

# %%
from pathlib import Path
from typing import Literal

import fsspec
import zarr
from filelock import BaseFileLock, FileLock
from zarr.errors import ContainsGroupError, GroupNotFoundError
from zarr.storage import DirectoryStore, FSStore, MemoryStore, Store, StoreLike

from ngio.utils import NgioFileExistsError, NgioFileNotFoundError, NgioValueError
from ngio.utils._errors import NgioError

AccessModeLiteral = Literal["r", "r+", "w", "w-", "a"]
# StoreLike is more restrictive than it could be
# but to make sure we can handle the store correctly
# we need to be more restrictive
NgioSupportedStore = (
    str | Path | fsspec.mapping.FSMap | FSStore | DirectoryStore | MemoryStore
)
GenericStore = Store | NgioSupportedStore
StoreOrGroup = GenericStore | zarr.Group


def _check_store(store) -> NgioSupportedStore:
    """Check the store and return a valid store."""
    if isinstance(store, NgioSupportedStore):
        return store

    raise NotImplementedError(
        f"Store type {type(store)} is not supported. "
        f"Supported types are: {NgioSupportedStore}"
    )


def _check_group(group: zarr.Group, mode: AccessModeLiteral) -> zarr.Group:
    """Check the group and return a valid group."""
    is_read_only = getattr(group, "_read_only", False)
    if is_read_only and mode in ["w", "w-"]:
        raise NgioValueError(
            "The group is read only. Cannot open in write mode ['w', 'w-']"
        )

    if mode == "r" and not is_read_only:
        # let's make sure we don't accidentally write to the group
        group = zarr.open_group(store=group.store, path=group.path, mode="r")

    return group


def open_group_wrapper(store: StoreOrGroup, mode: AccessModeLiteral) -> zarr.Group:
    """Wrapper around zarr.open_group with some additional checks.

    Args:
        store (StoreOrGroup): The store or group to open.
        mode (ReadOrEdirLiteral): The mode to open the group in.

    Returns:
        zarr.Group: The opened Zarr group.
    """
    if isinstance(store, zarr.Group):
        group = _check_group(store, mode)
        _check_store(group.store)
        return group

    try:
        _check_store(store)
        group = zarr.open_group(store=store, mode=mode)

    except ContainsGroupError as e:
        raise NgioFileExistsError(
            f"A Zarr group already exists at {store}, consider setting overwrite=True."
        ) from e

    except GroupNotFoundError as e:
        raise NgioFileNotFoundError(f"No Zarr group found at {store}") from e

    return group


class ZarrGroupHandler:
    """A simple wrapper around a Zarr group to handle metadata."""

    def __init__(
        self,
        store: StoreOrGroup,
        cache: bool = False,
        mode: AccessModeLiteral = "a",
        parallel_safe: bool = False,
        parent: "ZarrGroupHandler | None" = None,
    ):
        """Initialize the handler.

        Args:
            store (StoreOrGroup): The Zarr store or group containing the image data.
            meta_mode (str): The mode of the metadata handler.
            cache (bool): Whether to cache the metadata.
            mode (str): The mode of the store.
            parallel_safe (bool): If True, the handler will create a lock file to make
                that can be used to make the handler parallel safe.
                Be aware that the lock needs to be used manually.
            parent (ZarrGroupHandler | None): The parent handler.
        """
        if mode not in ["r", "r+", "w", "w-", "a"]:
            raise NgioValueError(f"Mode {mode} is not supported.")

        if parallel_safe and cache:
            raise NgioValueError(
                "The cache and parallel_safe options are mutually exclusive."
                "If you want to use the lock mechanism, you should not use the cache."
            )

        group = open_group_wrapper(store, mode)
        _store = group.store

        # Make sure the cache is set in the attrs
        # in the same way as the cache in the handler
        group.attrs.cache = cache

        if parallel_safe:
            if not isinstance(_store, DirectoryStore):
                raise NgioValueError(
                    "The store needs to be a DirectoryStore to use the lock mechanism. "
                    f"Instead, got {_store.__class__.__name__}."
                )
            store_path = Path(_store.path) / group.path
            self._lock_path = store_path.with_suffix(".lock")
            self._lock = FileLock(self._lock_path, timeout=10)

        else:
            self._lock_path = None
            self._lock = None

        self._group = group
        self._mode = mode
        self.use_cache = cache
        self._parallel_safe = parallel_safe
        self._cache = {}
        self._parent = parent

    def __repr__(self) -> str:
        """Return a string representation of the handler."""
        return (
            f"ZarrGroupHandler(full_url={self.full_url}, mode={self.mode}, "
            f"cache={self.use_cache}"
        )

    @property
    def store(self) -> StoreLike:
        """Return the store of the group."""
        return self.group.store

    @property
    def full_url(self) -> str | None:
        """Return the store path."""
        if isinstance(self.store, DirectoryStore | FSStore):
            _store_path = str(self.store.path)
            _store_path = _store_path.rstrip("/")
            return f"{self.store.path}/{self._group.path}"
        return None

    @property
    def mode(self) -> AccessModeLiteral:
        """Return the mode of the group."""
        return self._mode  # type: ignore

    @property
    def lock(self) -> BaseFileLock:
        """Return the lock."""
        if self._lock is None:
            raise NgioValueError(
                "The handler is not parallel safe. "
                "Reopen the handler with parallel_safe=True."
            )
        return self._lock

    @property
    def parent(self) -> "ZarrGroupHandler | None":
        """Return the parent handler."""
        return self._parent

    def remove_lock(self) -> None:
        """Return the lock."""
        if self._lock is None or self._lock_path is None:
            return None

        lock_path = Path(self._lock_path)
        if lock_path.exists() and self._lock.lock_counter == 0:
            lock_path.unlink()
            self._lock = None
            self._lock_path = None
            return None

        raise NgioValueError("The lock is still in use. Cannot remove it.")

    @property
    def group(self) -> zarr.Group:
        """Return the group."""
        return self._group

    def add_to_cache(self, key: str, value: object) -> None:
        """Add an object to the cache."""
        if not self.use_cache:
            return None
        self._cache[key] = value

    def get_from_cache(self, key: str) -> object | None:
        """Get an object from the cache."""
        if not self.use_cache:
            return None
        return self._cache.get(key, None)

    def clean_cache(self) -> None:
        """Clear the cached metadata."""
        self._cache = {}

    def load_attrs(self) -> dict:
        """Load the attributes of the group."""
        attrs = self.get_from_cache("attrs")
        if attrs is not None and isinstance(attrs, dict):
            return attrs

        attrs = dict(self.group.attrs)

        self.add_to_cache("attrs", attrs)
        return attrs

    def _write_attrs(self, attrs: dict, overwrite: bool = False) -> None:
        """Write the metadata to the store."""
        is_read_only = getattr(self._group, "_read_only", False)
        if is_read_only:
            raise NgioValueError("The group is read only. Cannot write metadata.")

        # we need to invalidate the current attrs cache
        self.add_to_cache("attrs", None)
        if overwrite:
            self.group.attrs.clear()

        self.group.attrs.update(attrs)

    def write_attrs(self, attrs: dict, overwrite: bool = False) -> None:
        """Write the metadata to the store."""
        # Maybe we should use the lock here
        self._write_attrs(attrs, overwrite)

    def _obj_get(self, path: str):
        """Get a group from the group."""
        group_or_array = self.get_from_cache(path)
        if group_or_array is not None:
            return group_or_array

        group_or_array = self.group.get(path, None)
        self.add_to_cache(path, group_or_array)
        return group_or_array

    def create_group(self, path: str, overwrite: bool = False) -> zarr.Group:
        """Create a group in the group."""
        if self.mode == "r":
            raise NgioValueError("Cannot create a group in read only mode.")

        try:
            group = self.group.create_group(path, overwrite=overwrite)
        except ContainsGroupError as e:
            raise NgioFileExistsError(
                f"A Zarr group already exists at {path}, "
                "consider setting overwrite=True."
            ) from e
        self.add_to_cache(path, group)
        return group

    def get_group(
        self,
        path: str,
        create_mode: bool = False,
        overwrite: bool = False,
    ) -> zarr.Group:
        """Get a group from the group.

        Args:
            path (str): The path to the group.
            create_mode (bool): If True, create the group if it does not exist.
            overwrite (bool): If True, overwrite the group if it exists.

        Returns:
            zarr.Group: The Zarr group.

        """
        if overwrite and not create_mode:
            raise NgioValueError("Cannot overwrite a group without create_mode=True.")

        if overwrite:
            return self.create_group(path, overwrite=overwrite)

        group = self._obj_get(path)
        if isinstance(group, zarr.Group):
            return group

        if group is not None:
            raise NgioValueError(
                f"The object at {path} is not a group, but a {type(group)}"
            )

        if not create_mode:
            raise NgioFileNotFoundError(f"No group found at {path}")
        group = self.create_group(path)
        return group

    def safe_get_group(
        self, path: str, create_mode: bool = False
    ) -> tuple[bool, zarr.Group | NgioError]:
        """Get a group from the group.

        Args:
            path (str): The path to the group.
            create_mode (bool): If True, create the group if it does not exist.

        Returns:
            zarr.Group | None: The Zarr group or None if it does not exist
                or an error occurs.

        """
        try:
            return True, self.get_group(path, create_mode)
        except NgioError as e:
            return False, e

    def get_array(self, path: str) -> zarr.Array:
        """Get an array from the group."""
        array = self._obj_get(path)
        if array is None:
            raise NgioFileNotFoundError(f"No array found at {path}")
        if not isinstance(array, zarr.Array):
            raise NgioValueError(
                f"The object at {path} is not an array, but a {type(array)}"
            )
        return array

    def create_array(
        self,
        path: str,
        shape: tuple[int, ...],
        dtype: str,
        chunks: tuple[int, ...] | None = None,
        overwrite: bool = False,
    ) -> zarr.Array:
        if self.mode == "r":
            raise NgioValueError("Cannot create an array in read only mode.")

        try:
            return self.group.zeros(
                name=path,
                shape=shape,
                dtype=dtype,
                chunks=chunks,
                dimension_separator="/",
                overwrite=overwrite,
            )
        except ContainsGroupError as e:
            raise NgioFileExistsError(
                f"A Zarr array already exists at {path}, "
                "consider setting overwrite=True."
            ) from e
        except Exception as e:
            raise NgioValueError(f"Error creating array at {path}") from e

    def derive_handler(
        self,
        path: str,
        overwrite: bool = False,
    ) -> "ZarrGroupHandler":
        """Derive a new handler from the current handler.

        Args:
            path (str): The path to the group.
            overwrite (bool): If True, overwrite the group if it exists.
        """
        group = self.get_group(path, create_mode=True, overwrite=overwrite)
        return ZarrGroupHandler(
            store=group,
            cache=self.use_cache,
            mode=self.mode,
            parallel_safe=self._parallel_safe,
            parent=self,
        )

    def safe_derive_handler(
        self,
        path: str,
        overwrite: bool = False,
    ) -> tuple[bool, "ZarrGroupHandler | NgioError"]:
        """Derive a new handler from the current handler."""
        try:
            return True, self.derive_handler(path, overwrite=overwrite)
        except NgioError as e:
            return False, e

    def copy_handler(self, handler: "ZarrGroupHandler") -> None:
        """Copy the group to a new store."""
        _, n_skipped, _ = zarr.copy_store(
            source=self.group.store,
            dest=handler.group.store,
            source_path=self.group.path,
            dest_path=handler.group.path,
            if_exists="replace",
        )
        if n_skipped > 0:
            raise NgioValueError(
                f"Error copying group to {handler.full_url}, "
                f"#{n_skipped} files where skipped."
            )


# %%



from pathlib import Path

import anndata as ad
import numpy as np
import pandas as pd
import pandas.api.types as ptypes
import pytest

from ngio.tables.backends import (
    ImplementedTableBackends,
    convert_anndata_to_pandas,
    convert_pandas_to_anndata,
)
from ngio.tables.backends._anndata_v1 import AnnDataBackend
from ngio.tables.backends._csv_v1 import CsvTableBackend
from ngio.tables.backends._json_v1 import JsonTableBackend
from ngio.utils import NgioValueError, ZarrGroupHandler


def test_backend_manager(tmp_path: Path):
    manager = ImplementedTableBackends()

    assert set(manager.available_backends) == {
        "experimental_json_v1",
        "anndata_v1",
        "experimental_csv_v1",
    }
    manager.add_backend(JsonTableBackend, overwrite=True)

    manager2 = ImplementedTableBackends()
    assert set(manager2.available_backends) == {
        "experimental_json_v1",
        "anndata_v1",
        "experimental_csv_v1",
    }
    assert set(manager.available_backends) == {
        "experimental_json_v1",
        "anndata_v1",
        "experimental_csv_v1",
    }

    store = tmp_path / "test_backend_manager.zarr"
    handler = ZarrGroupHandler(store=store, cache=True, mode="a")
    backend = manager.get_backend(
        backend_name="experimental_json_v1", group_handler=handler
    )
    assert isinstance(backend, JsonTableBackend)

    backend = manager.get_backend(None, handler)
    assert isinstance(backend, AnnDataBackend)

    with pytest.raises(NgioValueError):
        manager.get_backend("non_existent", handler)

    with pytest.raises(NgioValueError):
        manager.add_backend(JsonTableBackend)


def test_json_backend(tmp_path: Path):
    store = tmp_path / "test_json_backend.zarr"
    handler = ZarrGroupHandler(store=store, cache=True, mode="a")
    backend = JsonTableBackend(handler, index_type="str")

    assert backend.backend_name() == "experimental_json_v1"
    assert not backend.implements_anndata()
    assert backend.implements_pandas()

    test_table = pd.DataFrame(
        {"a": [1, 2, 3], "b": [4.0, 5.0, 6.0], "c": ["a", "b", "c"]}
    )
    test_table.index = test_table.index.astype(str)

    backend.write(test_table, metadata={"test": "test"})
    loaded_table = backend.load_as_pandas_df()

    assert loaded_table.equals(test_table)

    meta = backend._group_handler.load_attrs()
    assert meta["test"] == "test"
    assert meta["backend"] == "experimental_json_v1"

    a_data = backend.load_as_anndata()

    with pytest.raises(NotImplementedError):
        backend.write(a_data, metadata={"test": "test"})

    lf_data = backend.load_as_polars_lf()
    backend.write(lf_data, metadata={"test": "test"})


def test_csv_backend(tmp_path: Path):
    store = tmp_path / "test_csv_backend.zarr"
    handler = ZarrGroupHandler(store=store, cache=True, mode="a")
    backend = CsvTableBackend(handler)

    assert backend.backend_name() == "experimental_csv_v1"
    assert not backend.implements_anndata()
    assert backend.implements_pandas()

    test_table = pd.DataFrame(
        {"a": [1, 2, 3], "b": [4.0, 5.0, 6.0], "c": ["a", "b", "c"]}
    )

    backend.write(test_table, metadata={"test": "test"})
    loaded_table = backend.load_as_pandas_df()
    assert loaded_table.equals(test_table), loaded_table
    meta = backend._group_handler.load_attrs()
    assert meta["test"] == "test"
    assert meta["backend"] == "experimental_csv_v1"

    a_data = backend.load_as_anndata()
    with pytest.raises(NotImplementedError):
        backend.write(a_data, metadata={"test": "test"})

    lf_data = backend.load_as_polars_lf()
    backend.write(lf_data, metadata={"test": "test"})


def test_anndata_backend(tmp_path: Path):
    store = tmp_path / "test_anndata_backend.zarr"
    handler = ZarrGroupHandler(store=store, cache=True, mode="a")
    backend = AnnDataBackend(handler, index_type="int")

    assert backend.backend_name() == "anndata_v1"
    assert backend.implements_anndata()
    assert backend.implements_pandas()

    test_table = pd.DataFrame(
        {"a": [1, 2, 3], "b": [4.0, 5.0, 6.0], "c": ["a", "b", "c"]}
    )

    backend.write(test_table, metadata={"test": "test"})
    loaded_table = backend.load_as_pandas_df()

    for column in loaded_table.columns:
        # Since the transformation from anndata to dataframe is not perfect
        # We can only compare the columns
        pd.testing.assert_series_equal(loaded_table[column], test_table[column])

    meta = backend._group_handler.load_attrs()
    assert meta["test"] == "test"
    assert meta["backend"] == "anndata_v1"

    a_data = backend.load_as_anndata()
    backend.write(a_data, metadata={"test": "test"})

    lf_data = backend.load_as_polars_lf()
    backend.write(lf_data, metadata={"test": "test"})


@pytest.mark.parametrize(
    "index_label, index_type",
    [(None, "int"), ("label", "str"), ("label", "int")],
)
def test_anndata_to_dataframe(index_label: str | None, index_type: str):
    test_obs = pd.DataFrame({"a": [1, 2, 3], "c": ["a", "b", "c"]})
    test_x = pd.DataFrame({"b": [4.0, 5.0, 6.0]})

    if index_label is None:
        test_obs.index = test_obs.index.astype(str)
    else:
        test_obs.index = pd.Index(["1", "2", "3"], name=index_label)

    anndata = ad.AnnData(obs=test_obs, X=test_x)

    dataframe = convert_anndata_to_pandas(
        anndata,
        index_key=index_label,
        index_type=index_type,  # type: ignore[arg-type]
    )

    for column in test_obs.columns:
        pd.testing.assert_series_equal(
            dataframe[column], test_obs[column], check_index=False
        )

    for column in test_x.columns:
        pd.testing.assert_series_equal(
            dataframe[column], test_x[column], check_index=False
        )

    if index_label is not None:
        assert dataframe.index.name == index_label
        if index_type == "int":
            assert ptypes.is_integer_dtype(dataframe.index)
        elif index_type == "str":
            assert ptypes.is_string_dtype(dataframe.index)


@pytest.mark.parametrize(
    "index_label, index_type",
    [(None, "int"), ("label", "str"), ("label", "int")],
)
def test_dataframe_to_anndata(index_label: str | None, index_type: str):
    test_table = pd.DataFrame(
        {
            "a": [1, 2, 3],
            "b": [4.0, 5.0, 6.0],
            "c": ["a", "b", "c"],
            "d": [4.0, 5.0, 6.0],
        }
    )

    if index_label is not None and index_type == "int":
        test_table.index = pd.Index([1, 2, 3], name=index_label)
    elif index_label is not None and index_type == "str":
        test_table.index = pd.Index(["a", "b", "c"], name=index_label)

    anndata = convert_pandas_to_anndata(
        test_table,
        index_key=index_label,
    )

    for column in anndata.obs.columns:
        pd.testing.assert_series_equal(
            anndata.obs[column], test_table[column], check_index=False
        )

    if index_label is not None:
        assert anndata.obs.index.name == index_label

    for i, column in enumerate(anndata.var.index):
        np.testing.assert_allclose(anndata.X[:, i], test_table[column].values)  # type: ignore


@pytest.mark.parametrize(
    "index_label, index_type",
    [(None, "int"), ("label", "str"), ("label", "int")],
)
def test_round_trip(index_label: str | None, index_type: str):
    test_table = pd.DataFrame(
        {
            "a": [1, 2, 3],
            "b": [4.0, 5.0, 6.0],
            "c": ["a", "b", "c"],
            "d": [4.0, 5.0, 6.0],
        }
    )

    if index_label is not None and index_type == "int":
        test_table.index = pd.Index([1, 2, 3], name=index_label)
    elif index_label is not None and index_type == "str":
        test_table.index = pd.Index(["a", "b", "c"], name=index_label)

    anndata = convert_pandas_to_anndata(
        test_table,
        index_key=index_label,
    )
    datafame = convert_anndata_to_pandas(
        anndata,
        index_key=index_label,
        index_type=index_type,  # type: ignore[arg-type]
    )

    for column in datafame.columns:
        pd.testing.assert_series_equal(
            datafame[column], test_table[column], check_index=True
        )



"""Fractal internal module for axes handling."""

from collections.abc import Collection
from enum import Enum
from logging import Logger
from typing import Literal, TypeVar

import numpy as np
from pydantic import BaseModel, ConfigDict, Field

from ngio.utils import NgioValidationError, NgioValueError

logger = Logger(__name__)

T = TypeVar("T")

################################################################################################
#
# Axis Types and Units
# We define a small set of axis types and units that can be used in the metadata.
# This axis types are more restrictive than the OME standard.
# We do that to simplify the data processing.
#
#################################################################################################


class AxisType(str, Enum):
    """Allowed axis types."""

    channel = "channel"
    time = "time"
    space = "space"


SpaceUnits = Literal[
    "micrometer",
    "nanometer",
    "angstrom",
    "picometer",
    "millimeter",
    "centimeter",
    "decimeter",
    "meter",
    "inch",
    "foot",
    "yard",
    "mile",
    "kilometer",
    "hectometer",
    "megameter",
    "gigameter",
    "terameter",
    "petameter",
    "exameter",
    "parsec",
    "femtometer",
    "attometer",
    "zeptometer",
    "yoctometer",
    "zettameter",
    "yottameter",
]
DefaultSpaceUnit = "micrometer"

TimeUnits = Literal[
    "attosecond",
    "centisecond",
    "day",
    "decisecond",
    "exasecond",
    "femtosecond",
    "gigasecond",
    "hectosecond",
    "hour",
    "kilosecond",
    "megasecond",
    "microsecond",
    "millisecond",
    "minute",
    "nanosecond",
    "petasecond",
    "picosecond",
    "second",
    "terasecond",
    "yoctosecond",
    "yottasecond",
    "zeptosecond",
    "zettasecond",
]
DefaultTimeUnit = "second"


class Axis(BaseModel):
    """Axis infos model."""

    on_disk_name: str
    unit: str | None = None
    axis_type: AxisType | None = None

    model_config = ConfigDict(extra="forbid", frozen=True)

    def implicit_type_cast(self, cast_type: AxisType) -> "Axis":
        unit = self.unit
        if self.axis_type != cast_type:
            logger.warning(
                f"Axis {self.on_disk_name} has type {self.axis_type}. "
                f"Casting to {cast_type}."
            )

        if cast_type == AxisType.time and unit is None:
            logger.warning(
                f"Time axis {self.on_disk_name} has unit {self.unit}. "
                f"Casting to {DefaultSpaceUnit}."
            )
            unit = DefaultTimeUnit

        if cast_type == AxisType.space and unit is None:
            logger.warning(
                f"Space axis {self.on_disk_name} has unit {unit}. "
                f"Casting to {DefaultSpaceUnit}."
            )
            unit = DefaultSpaceUnit

        return Axis(on_disk_name=self.on_disk_name, axis_type=cast_type, unit=unit)

    def canonical_axis_cast(self, canonical_name: str) -> "Axis":
        """Cast the implicit axis to the correct type."""
        match canonical_name:
            case "t":
                if self.axis_type != AxisType.time or self.unit is None:
                    return self.implicit_type_cast(AxisType.time)
            case "c":
                if self.axis_type != AxisType.channel:
                    return self.implicit_type_cast(AxisType.channel)
            case "z" | "y" | "x":
                if self.axis_type != AxisType.space or self.unit is None:
                    return self.implicit_type_cast(AxisType.space)
        return self


################################################################################################
#
# Axes Handling
# We define a unique mapping to match the axes on disk to the canonical axes.
# The canonical axes are the ones that are used consistently in the NGIO internal API.
# The canonical axes ordered are: t, c, z, y, x.
#
#################################################################################################


def canonical_axes_order() -> tuple[str, str, str, str, str]:
    """Get the canonical axes order."""
    return "t", "c", "z", "y", "x"


def canonical_label_axes_order() -> tuple[str, str, str, str]:
    """Get the canonical axes order."""
    return "t", "z", "y", "x"


class AxesSetup(BaseModel):
    """Axes setup model.

    This model is used to map the on disk axes to the canonical OME-Zarr axes.
    """

    x: str = "x"
    y: str = "y"
    z: str = "z"
    c: str = "c"
    t: str = "t"
    others: list[str] = Field(default_factory=list)

    model_config = ConfigDict(extra="forbid", frozen=True)


def _check_unique_names(axes: Collection[Axis]):
    """Check if all axes on disk have unique names."""
    names = [ax.on_disk_name for ax in axes]
    if len(set(names)) != len(names):
        duplicates = {item for item in names if names.count(item) > 1}
        raise NgioValidationError(
            f"All axes must be unique. But found duplicates axes {duplicates}"
        )


def _check_non_canonical_axes(axes_setup: AxesSetup, allow_non_canonical_axes: bool):
    """Check if all axes are known."""
    if not allow_non_canonical_axes and len(axes_setup.others) > 0:
        raise NgioValidationError(
            f"Unknown axes {axes_setup.others}. Please set "
            "`allow_non_canonical_axes=True` to ignore them"
        )


def _check_axes_validity(axes: Collection[Axis], axes_setup: AxesSetup):
    """Check if all axes are valid."""
    _axes_setup = axes_setup.model_dump(exclude={"others"})
    _all_known_axes = [*_axes_setup.values(), *axes_setup.others]
    for ax in axes:
        if ax.on_disk_name not in _all_known_axes:
            raise NgioValidationError(
                f"Invalid axis name '{ax.on_disk_name}'. "
                f"Please correct map `{ax.on_disk_name}` "
                f"using the AxesSetup model {axes_setup}"
            )


def _check_canonical_order(
    axes: Collection[Axis], axes_setup: AxesSetup, strict_canonical_order: bool
):
    """Check if the axes are in the canonical order."""
    if not strict_canonical_order:
        return
    _on_disk_names = [ax.on_disk_name for ax in axes]
    _canonical_order = []
    for name in canonical_axes_order():
        mapped_name = getattr(axes_setup, name)
        if mapped_name in _on_disk_names:
            _canonical_order.append(mapped_name)

    if _on_disk_names != _canonical_order:
        raise NgioValidationError(
            f"Invalid axes order. The axes must be in the canonical order. "
            f"Expected {_canonical_order}, but found {_on_disk_names}"
        )


def validate_axes(
    axes: Collection[Axis],
    axes_setup: AxesSetup,
    allow_non_canonical_axes: bool = False,
    strict_canonical_order: bool = False,
) -> None:
    """Validate the axes."""
    if allow_non_canonical_axes and strict_canonical_order:
        raise NgioValidationError(
            "`allow_non_canonical_axes` and"
            "`strict_canonical_order` cannot be true at the same time."
            "If non canonical axes are allowed, the order cannot be checked."
        )
    _check_unique_names(axes=axes)
    _check_non_canonical_axes(
        axes_setup=axes_setup, allow_non_canonical_axes=allow_non_canonical_axes
    )
    _check_axes_validity(axes=axes, axes_setup=axes_setup)
    _check_canonical_order(
        axes=axes, axes_setup=axes_setup, strict_canonical_order=strict_canonical_order
    )


class AxesTransformation(BaseModel):
    model_config = ConfigDict(extra="forbid", frozen=True, arbitrary_types_allowed=True)


class AxesTranspose(AxesTransformation):
    axes: tuple[int, ...]


class AxesExpand(AxesTransformation):
    axes: tuple[int, ...]


class AxesSqueeze(AxesTransformation):
    axes: tuple[int, ...]


class AxesMapper:
    """Map on disk axes to canonical axes.

    This class is used to map the on disk axes to the canonical axes.

    """

    def __init__(
        self,
        # spec dictated args
        on_disk_axes: Collection[Axis],
        # user defined args
        axes_setup: AxesSetup | None = None,
        allow_non_canonical_axes: bool = False,
        strict_canonical_order: bool = False,
    ):
        """Create a new AxesMapper object.

        Args:
            on_disk_axes (list[Axis]): The axes on disk.
            axes_setup (AxesSetup, optional): The axis setup. Defaults to None.
            allow_non_canonical_axes (bool, optional): Allow non canonical axes.
            strict_canonical_order (bool, optional): Check if the axes are in the
                canonical order. Defaults to False.
        """
        axes_setup = axes_setup if axes_setup is not None else AxesSetup()

        validate_axes(
            axes=on_disk_axes,
            axes_setup=axes_setup,
            allow_non_canonical_axes=allow_non_canonical_axes,
            strict_canonical_order=strict_canonical_order,
        )

        self._allow_non_canonical_axes = allow_non_canonical_axes
        self._strict_canonical_order = strict_canonical_order

        self._canonical_order = canonical_axes_order()
        self._extended_canonical_order = [*axes_setup.others, *self._canonical_order]

        self._on_disk_axes = on_disk_axes
        self._axes_setup = axes_setup

        self._name_mapping = self._compute_name_mapping()
        self._index_mapping = self._compute_index_mapping()

        # Validate the axes type and cast them if necessary
        # This needs to be done after the name mapping is computed
        self.validate_axex_type()

    def _compute_name_mapping(self):
        """Compute the name mapping.

        The name mapping is a dictionary with keys the canonical axes names
        and values the on disk axes names.
        """
        _name_mapping = {}
        axis_setup_dict = self._axes_setup.model_dump(exclude={"others"})
        _on_disk_names = self.on_disk_axes_names
        for canonical_key, on_disk_value in axis_setup_dict.items():
            if on_disk_value in _on_disk_names:
                _name_mapping[canonical_key] = on_disk_value
            else:
                _name_mapping[canonical_key] = None

        for on_disk_name in _on_disk_names:
            if on_disk_name not in _name_mapping.keys():
                _name_mapping[on_disk_name] = on_disk_name

        for other in self._axes_setup.others:
            if other not in _name_mapping.keys():
                _name_mapping[other] = None
        return _name_mapping

    def _compute_index_mapping(self):
        """Compute the index mapping.

        The index mapping is a dictionary with keys the canonical axes names
        and values the on disk axes index.
        """
        _index_mapping = {}
        for canonical_key, on_disk_value in self._name_mapping.items():
            if on_disk_value is not None:
                _index_mapping[canonical_key] = self.on_disk_axes_names.index(
                    on_disk_value
                )
            else:
                _index_mapping[canonical_key] = None
        return _index_mapping

    @property
    def axes_setup(self) -> AxesSetup:
        """Return the axes setup."""
        return self._axes_setup

    @property
    def on_disk_axes(self) -> list[Axis]:
        return list(self._on_disk_axes)

    @property
    def on_disk_axes_names(self) -> list[str]:
        return [ax.on_disk_name for ax in self._on_disk_axes]

    @property
    def allow_non_canonical_axes(self) -> bool:
        """Return if non canonical axes are allowed."""
        return self._allow_non_canonical_axes

    @property
    def strict_canonical_order(self) -> bool:
        """Return if strict canonical order is enforced."""
        return self._strict_canonical_order

    def get_index(self, name: str) -> int | None:
        """Get the index of the axis by name."""
        if name not in self._index_mapping.keys():
            raise NgioValueError(
                f"Invalid axis name '{name}'. "
                f"Possible values are {self._index_mapping.keys()}"
            )
        return self._index_mapping[name]

    def get_axis(self, name: str) -> Axis | None:
        """Get the axis object by name."""
        index = self.get_index(name)
        if index is None:
            return None
        return self.on_disk_axes[index]

    def validate_axex_type(self):
        """Validate the axes type.

        If the axes type is not correct, a warning is issued.
        and the axis is implicitly cast to the correct type.
        """
        new_axes = []
        for axes in self.on_disk_axes:
            for name in self._canonical_order:
                if axes == self.get_axis(name):
                    new_axes.append(axes.canonical_axis_cast(name))
                    break
            else:
                new_axes.append(axes)
        self._on_disk_axes = new_axes

    def _change_order(
        self, names: Collection[str]
    ) -> tuple[tuple[int, ...], tuple[int, ...]]:
        unique_names = set()
        for name in names:
            if name not in self._index_mapping.keys():
                raise NgioValueError(
                    f"Invalid axis name '{name}'. "
                    f"Possible values are {self._index_mapping.keys()}"
                )
            _unique_name = self._name_mapping.get(name)
            if _unique_name is None:
                continue
            if _unique_name in unique_names:
                raise NgioValueError(
                    f"Duplicate axis name, two or more '{_unique_name}' were found. "
                    f"Please provide unique names."
                )
            unique_names.add(_unique_name)

        if len(self.on_disk_axes_names) > len(unique_names):
            missing_names = set(self.on_disk_axes_names) - unique_names
            raise NgioValueError(
                f"Some axes where not queried. "
                f"Please provide the following missing axes {missing_names}"
            )
        _indices, _insert = [], []
        for i, name in enumerate(names):
            _index = self._index_mapping[name]
            if _index is None:
                _insert.append(i)
            else:
                _indices.append(self._index_mapping[name])
        return tuple(_indices), tuple(_insert)

    def to_order(self, names: Collection[str]) -> tuple[AxesTransformation, ...]:
        """Get the new order of the axes."""
        _indices, _insert = self._change_order(names)
        return AxesTranspose(axes=_indices), AxesExpand(axes=_insert)

    def from_order(self, names: Collection[str]) -> tuple[AxesTransformation, ...]:
        """Get the new order of the axes."""
        _indices, _insert = self._change_order(names)
        # Inverse transpose is just the transpose with the inverse indices
        _reverse_indices = tuple(np.argsort(_indices))
        return AxesSqueeze(axes=_insert), AxesTranspose(axes=_reverse_indices)

    def to_canonical(self) -> tuple[AxesTransformation, ...]:
        """Get the new order of the axes."""
        return self.to_order(self._extended_canonical_order)

    def from_canonical(self) -> tuple[AxesTransformation, ...]:
        """Get the new order of the axes."""
        return self.from_order(self._extended_canonical_order)


def canonical_axes(
    axes_names: Collection[str],
    space_units: SpaceUnits | None = DefaultSpaceUnit,
    time_units: TimeUnits | None = DefaultTimeUnit,
) -> list[Axis]:
    """Create a new canonical axes mapper.

    Args:
        axes_names (Collection[str] | int): The axes names on disk.
            - The axes should be in ['t', 'c', 'z', 'y', 'x']
            - The axes should be in strict canonical order.
            - If an integer is provided, the axes are created from the last axis
              to the first
                e.g. 3 -> ["z", "y", "x"]
        space_units (SpaceUnits, optional): The space units. Defaults to None.
        time_units (TimeUnits, optional): The time units. Defaults to None.

    """
    axes = []
    for name in axes_names:
        match name:
            case "t":
                axes.append(
                    Axis(on_disk_name=name, axis_type=AxisType.time, unit=time_units)
                )
            case "c":
                axes.append(Axis(on_disk_name=name, axis_type=AxisType.channel))
            case "z" | "y" | "x":
                axes.append(
                    Axis(on_disk_name=name, axis_type=AxisType.space, unit=space_units)
                )
            case _:
                raise NgioValueError(
                    f"Invalid axis name '{name}'. "
                    "Only 't', 'c', 'z', 'y', 'x' are allowed."
                )

    return axes



"""Generic class to handle Image-like data in a OME-NGFF file."""

from collections.abc import Collection
from typing import Literal

from dask import array as da

from ngio.common import Dimensions
from ngio.images.abstract_image import AbstractImage, consolidate_image
from ngio.images.create import create_empty_image_container
from ngio.ome_zarr_meta import (
    ImageMetaHandler,
    NgioImageMeta,
    PixelSize,
    find_image_meta_handler,
)
from ngio.ome_zarr_meta.ngio_specs import (
    Channel,
    ChannelsMeta,
    ChannelVisualisation,
    DefaultSpaceUnit,
    DefaultTimeUnit,
    SpaceUnits,
    TimeUnits,
)
from ngio.utils import (
    NgioValidationError,
    StoreOrGroup,
    ZarrGroupHandler,
)


def _check_channel_meta(meta: NgioImageMeta, dimension: Dimensions) -> ChannelsMeta:
    """Check the channel metadata."""
    c_dim = dimension.get("c", strict=False)
    c_dim = 1 if c_dim is None else c_dim

    if meta.channels_meta is None:
        return ChannelsMeta.default_init(labels=c_dim)

    if len(meta.channels) != c_dim:
        raise NgioValidationError(
            "The number of channels does not match the image. "
            f"Expected {len(meta.channels)} channels, got {c_dim}."
        )

    return meta.channels_meta


class Image(AbstractImage[ImageMetaHandler]):
    """A class to handle a single image (or level) in an OME-Zarr image.

    This class is meant to be subclassed by specific image types.
    """

    def __init__(
        self,
        group_handler: ZarrGroupHandler,
        path: str,
        meta_handler: ImageMetaHandler | None,
    ) -> None:
        """Initialize the Image at a single level.

        Args:
            group_handler: The Zarr group handler.
            path: The path to the image in the ome_zarr file.
            meta_handler: The image metadata handler.

        """
        if meta_handler is None:
            meta_handler = find_image_meta_handler(group_handler)
        super().__init__(
            group_handler=group_handler, path=path, meta_handler=meta_handler
        )
        self._channels_meta = _check_channel_meta(self.meta, self.dimensions)

    @property
    def meta(self) -> NgioImageMeta:
        """Return the metadata."""
        return self._meta_handler.meta

    @property
    def channel_labels(self) -> list[str]:
        """Return the channels of the image."""
        channel_labels = []
        for c in self._channels_meta.channels:
            channel_labels.append(c.label)
        return channel_labels

    @property
    def wavelength_ids(self) -> list[str | None]:
        """Return the list of wavelength of the image."""
        wavelength_ids = []
        for c in self._channels_meta.channels:
            wavelength_ids.append(c.wavelength_id)
        return wavelength_ids

    @property
    def num_channels(self) -> int:
        """Return the number of channels."""
        return len(self._channels_meta.channels)

    def consolidate(
        self,
        order: Literal[0, 1, 2] = 1,
        mode: Literal["dask", "numpy", "coarsen"] = "dask",
    ) -> None:
        """Consolidate the label on disk."""
        consolidate_image(self, order=order, mode=mode)


class ImagesContainer:
    """A class to handle the /labels group in an OME-NGFF file."""

    def __init__(self, group_handler: ZarrGroupHandler) -> None:
        """Initialize the LabelGroupHandler."""
        self._group_handler = group_handler
        self._meta_handler = find_image_meta_handler(group_handler)

    @property
    def meta(self) -> NgioImageMeta:
        """Return the metadata."""
        return self._meta_handler.meta

    @property
    def levels(self) -> int:
        """Return the number of levels in the image."""
        return self._meta_handler.meta.levels

    @property
    def levels_paths(self) -> list[str]:
        """Return the paths of the levels in the image."""
        return self._meta_handler.meta.paths

    @property
    def num_channels(self) -> int:
        """Return the number of channels."""
        image = self.get()
        return image.num_channels

    @property
    def channel_labels(self) -> list[str]:
        """Return the channels of the image."""
        image = self.get()
        return image.channel_labels

    @property
    def wavelength_ids(self) -> list[str | None]:
        """Return the wavelength of the image."""
        image = self.get()
        return image.wavelength_ids

    def set_channel_meta(
        self,
        labels: Collection[str] | int | None = None,
        wavelength_id: Collection[str] | None = None,
        start: Collection[float] | None = None,
        end: Collection[float] | None = None,
        percentiles: tuple[float, float] | None = None,
        colors: Collection[str] | None = None,
        active: Collection[bool] | None = None,
        **omero_kwargs: dict,
    ) -> None:
        """Create a ChannelsMeta object with the default unit.

        Args:
            labels(Collection[str] | int): The list of channels names in the image.
                If an integer is provided, the channels will be named "channel_i".
            wavelength_id(Collection[str] | None): The wavelength ID of the channel.
                If None, the wavelength ID will be the same as the channel name.
            start(Collection[float] | None): The start value for each channel.
                If None, the start value will be computed from the image.
            end(Collection[float] | None): The end value for each channel.
                If None, the end value will be computed from the image.
            percentiles(tuple[float, float] | None): The start and end percentiles
                for each channel. If None, the percentiles will not be computed.
            colors(Collection[str, NgioColors] | None): The list of colors for the
                channels. If None, the colors will be random.
            active (Collection[bool] | None):active(bool): Whether the channel should
                be shown by default.
            omero_kwargs(dict): Extra fields to store in the omero attributes.
        """
        low_res_dataset = self.meta.get_lowest_resolution_dataset()
        ref_image = self.get(path=low_res_dataset.path)

        if start is not None and end is None:
            raise NgioValidationError(
                "If start is provided, end must be provided as well."
            )
        if end is not None and start is None:
            raise NgioValidationError(
                "If end is provided, start must be provided as well."
            )

        if start is not None and percentiles is not None:
            raise NgioValidationError(
                "If start and end are provided, percentiles must be None."
            )

        if percentiles is not None:
            start, end = compute_image_percentile(
                ref_image,
                start_percentile=percentiles[0],
                end_percentile=percentiles[1],
            )
        elif start is not None and end is not None:
            if len(start) != len(end):
                raise NgioValidationError(
                    "The start and end lists must have the same length."
                )
            if len(start) != self.num_channels:
                raise NgioValidationError(
                    "The start and end lists must have the same length as "
                    "the number of channels."
                )

            start = list(start)
            end = list(end)

        else:
            start, end = None, None

        if labels is None:
            labels = ref_image.num_channels

        channel_meta = ChannelsMeta.default_init(
            labels=labels,
            wavelength_id=wavelength_id,
            colors=colors,
            start=start,
            end=end,
            active=active,
            data_type=ref_image.dtype,
            **omero_kwargs,
        )

        meta = self.meta
        meta.set_channels_meta(channel_meta)
        self._meta_handler.write_meta(meta)

    def set_channel_percentiles(
        self,
        start_percentile: float = 0.1,
        end_percentile: float = 99.9,
    ) -> None:
        """Update the percentiles of the channels."""
        if self.meta._channels_meta is None:
            raise NgioValidationError("The channels meta is not initialized.")

        low_res_dataset = self.meta.get_lowest_resolution_dataset()
        ref_image = self.get(path=low_res_dataset.path)
        starts, ends = compute_image_percentile(
            ref_image, start_percentile=start_percentile, end_percentile=end_percentile
        )

        channels = []
        for c, channel in enumerate(self.meta._channels_meta.channels):
            new_v = ChannelVisualisation(
                start=starts[c],
                end=ends[c],
                **channel.channel_visualisation.model_dump(exclude={"start", "end"}),
            )
            new_c = Channel(
                channel_visualisation=new_v,
                **channel.model_dump(exclude={"channel_visualisation"}),
            )
            channels.append(new_c)

        new_meta = ChannelsMeta(channels=channels)

        meta = self.meta
        meta.set_channels_meta(new_meta)
        self._meta_handler.write_meta(meta)

    def set_axes_unit(
        self,
        space_unit: SpaceUnits = DefaultSpaceUnit,
        time_unit: TimeUnits = DefaultTimeUnit,
    ) -> None:
        """Set the axes unit of the image.

        Args:
            space_unit (SpaceUnits): The space unit of the image.
            time_unit (TimeUnits): The time unit of the image.
        """
        meta = self.meta
        meta = meta.to_units(space_unit=space_unit, time_unit=time_unit)
        self._meta_handler.write_meta(meta)

    def derive(
        self,
        store: StoreOrGroup,
        ref_path: str | None = None,
        shape: Collection[int] | None = None,
        labels: Collection[str] | None = None,
        pixel_size: PixelSize | None = None,
        axes_names: Collection[str] | None = None,
        name: str | None = None,
        chunks: Collection[int] | None = None,
        dtype: str | None = None,
        overwrite: bool = False,
    ) -> "ImagesContainer":
        """Create an empty OME-Zarr image from an existing image.

        Args:
            store (StoreOrGroup): The Zarr store or group to create the image in.
            ref_path (str | None): The path to the reference image in
                the image container.
            shape (Collection[int] | None): The shape of the new image.
            labels (Collection[str] | None): The labels of the new image.
            pixel_size (PixelSize | None): The pixel size of the new image.
            axes_names (Collection[str] | None): The axes names of the new image.
            name (str | None): The name of the new image.
            chunks (Collection[int] | None): The chunk shape of the new image.
            dtype (str | None): The data type of the new image.
            overwrite (bool): Whether to overwrite an existing image.

        Returns:
            ImagesContainer: The new image
        """
        return derive_image_container(
            image_container=self,
            store=store,
            ref_path=ref_path,
            shape=shape,
            labels=labels,
            pixel_size=pixel_size,
            axes_names=axes_names,
            name=name,
            chunks=chunks,
            dtype=dtype,
            overwrite=overwrite,
        )

    def get(
        self,
        path: str | None = None,
        pixel_size: PixelSize | None = None,
        strict: bool = False,
    ) -> Image:
        """Get an image at a specific level.

        Args:
            path (str | None): The path to the image in the ome_zarr file.
            pixel_size (PixelSize | None): The pixel size of the image.
            strict (bool): Only used if the pixel size is provided. If True, the
                pixel size must match the image pixel size exactly. If False, the
                closest pixel size level will be returned.

        """
        dataset = self._meta_handler.meta.get_dataset(
            path=path, pixel_size=pixel_size, strict=strict
        )
        return Image(
            group_handler=self._group_handler,
            path=dataset.path,
            meta_handler=self._meta_handler,
        )


def compute_image_percentile(
    image: Image,
    start_percentile: float = 0.1,
    end_percentile: float = 99.9,
) -> tuple[list[float], list[float]]:
    """Compute the start and end percentiles for each channel of an image.

    Args:
        image: The image to compute the percentiles for.
        start_percentile: The start percentile to compute.
        end_percentile: The end percentile to compute.

    Returns:
        A tuple containing the start and end percentiles for each channel.
    """
    starts, ends = [], []
    for c in range(image.num_channels):
        if image.num_channels == 1:
            data = image.get_array(mode="dask").ravel()
        else:
            data = image.get_array(c=c, mode="dask").ravel()
        # remove all the zeros
        mask = data > 1e-16
        data = data[mask]
        _data = data.compute()
        if _data.size == 0:
            starts.append(0.0)
            ends.append(0.0)
            continue

        # compute the percentiles
        _s_perc, _e_perc = da.percentile(
            data, [start_percentile, end_percentile], method="nearest"
        ).compute()

        starts.append(float(_s_perc))
        ends.append(float(_e_perc))
    return starts, ends


def derive_image_container(
    image_container: ImagesContainer,
    store: StoreOrGroup,
    ref_path: str | None = None,
    shape: Collection[int] | None = None,
    labels: Collection[str] | None = None,
    pixel_size: PixelSize | None = None,
    axes_names: Collection[str] | None = None,
    name: str | None = None,
    chunks: Collection[int] | None = None,
    dtype: str | None = None,
    overwrite: bool = False,
) -> ImagesContainer:
    """Create an empty OME-Zarr image from an existing image.

    Args:
        image_container (ImagesContainer): The image container to derive the new image.
        store (StoreOrGroup): The Zarr store or group to create the image in.
        ref_path (str | None): The path to the reference image in the image container.
        shape (Collection[int] | None): The shape of the new image.
        labels (Collection[str] | None): The labels of the new image.
        pixel_size (PixelSize | None): The pixel size of the new image.
        axes_names (Collection[str] | None): The axes names of the new image.
        name (str | None): The name of the new image.
        chunks (Collection[int] | None): The chunk shape of the new image.
        dtype (str | None): The data type of the new image.
        overwrite (bool): Whether to overwrite an existing image.

    Returns:
        ImagesContainer: The new image

    """
    if ref_path is None:
        ref_image = image_container.get()
    else:
        ref_image = image_container.get(path=ref_path)

    ref_meta = ref_image.meta

    if shape is None:
        shape = ref_image.shape

    if pixel_size is None:
        pixel_size = ref_image.pixel_size

    if axes_names is None:
        axes_names = ref_meta.axes_mapper.on_disk_axes_names

    if len(axes_names) != len(shape):
        raise NgioValidationError(
            "The axes names of the new image does not match the reference image."
            f"Got {axes_names} for shape {shape}."
        )

    if chunks is None:
        chunks = ref_image.chunks

    if len(chunks) != len(shape):
        raise NgioValidationError(
            "The chunks of the new image does not match the reference image."
            f"Got {chunks} for shape {shape}."
        )

    if name is None:
        name = ref_meta.name

    if dtype is None:
        dtype = ref_image.dtype
    handler = create_empty_image_container(
        store=store,
        shape=shape,
        pixelsize=pixel_size.x,
        z_spacing=pixel_size.z,
        time_spacing=pixel_size.t,
        levels=ref_meta.levels,
        yx_scaling_factor=ref_meta.yx_scaling(),
        z_scaling_factor=ref_meta.z_scaling(),
        time_unit=pixel_size.time_unit,
        space_unit=pixel_size.space_unit,
        axes_names=axes_names,
        name=name,
        chunks=chunks,
        dtype=dtype,
        overwrite=overwrite,
        version=ref_meta.version,
    )
    image_container = ImagesContainer(handler)

    if ref_image.num_channels == image_container.num_channels:
        _labels = ref_image.channel_labels
        wavelength_id = ref_image.wavelength_ids

        colors = [
            c.channel_visualisation.color for c in ref_image._channels_meta.channels
        ]
        active = [
            c.channel_visualisation.active for c in ref_image._channels_meta.channels
        ]
        start = [
            c.channel_visualisation.start for c in ref_image._channels_meta.channels
        ]
        end = [c.channel_visualisation.end for c in ref_image._channels_meta.channels]
    else:
        _labels = None
        wavelength_id = None
        colors = None
        active = None
        start = None
        end = None

    if labels is not None:
        if len(labels) != image_container.num_channels:
            raise NgioValidationError(
                "The number of labels does not match the number of channels."
            )
        _labels = labels

    image_container.set_channel_meta(
        labels=_labels,
        wavelength_id=wavelength_id,
        percentiles=None,
        colors=colors,
        active=active,
        start=start,
        end=end,
    )
    return image_container



from pathlib import Path

import numpy as np
import pytest

from ngio import create_empty_ome_zarr, create_ome_zarr_from_array
from ngio.utils import NgioValueError


@pytest.mark.parametrize(
    "create_kwargs",
    [
        {
            "store": "test_image_yx._zarr",
            "shape": (64, 64),
            "xy_pixelsize": 0.5,
            "axes_names": ["y", "x"],
        },
        {
            "store": "test_image_cyx.zarr",
            "shape": (2, 64, 64),
            "xy_pixelsize": 0.5,
            "axes_names": ["c", "y", "x"],
            "channel_labels": ["channel1", "channel2"],
        },
        {
            "store": "test_image_zyx.zarr",
            "shape": (3, 64, 64),
            "xy_pixelsize": 0.5,
            "z_spacing": 2.0,
            "axes_names": ["z", "y", "x"],
        },
        {
            "store": "test_image_czyx.zarr",
            "shape": (2, 3, 64, 64),
            "xy_pixelsize": 0.5,
            "z_spacing": 2.0,
            "axes_names": ["c", "z", "y", "x"],
            "channel_labels": ["channel1", "channel2"],
        },
        {
            "store": "test_image_c1yx.zarr",
            "shape": (2, 1, 64, 64),
            "xy_pixelsize": 0.5,
            "z_spacing": 1.0,
            "axes_names": ["c", "z", "y", "x"],
            "channel_labels": ["channel1", "channel2"],
        },
        {
            "store": "test_image_tyx.zarr",
            "shape": (4, 64, 64),
            "xy_pixelsize": 0.5,
            "time_spacing": 4.0,
            "axes_names": ["t", "y", "x"],
        },
        {
            "store": "test_image_tcyx.zarr",
            "shape": (4, 2, 64, 64),
            "xy_pixelsize": 0.5,
            "time_spacing": 4.0,
            "axes_names": ["t", "c", "y", "x"],
            "channel_labels": ["channel1", "channel2"],
        },
        {
            "store": "test_image_tzyx.zarr",
            "shape": (4, 3, 64, 64),
            "xy_pixelsize": 0.5,
            "z_spacing": 2.0,
            "time_spacing": 4.0,
            "axes_names": ["t", "z", "y", "x"],
        },
        {
            "store": "test_image_tczyx.zarr",
            "shape": (4, 2, 3, 64, 64),
            "xy_pixelsize": 0.5,
            "z_spacing": 2.0,
            "time_spacing": 4.0,
            "axes_names": ["t", "c", "z", "y", "x"],
            "channel_labels": ["channel1", "channel2"],
        },
    ],
)
def test_create_empty(tmp_path: Path, create_kwargs: dict):
    create_kwargs["store"] = tmp_path / create_kwargs["store"]
    ome_zarr = create_empty_ome_zarr(**create_kwargs, dtype="uint8", levels=1)
    ome_zarr.derive_label("label1")

    shape = create_kwargs.pop("shape")
    array = np.random.randint(0, 255, shape, dtype="uint8")
    create_ome_zarr_from_array(array=array, **create_kwargs, levels=1, overwrite=True)


def test_create_fail(tmp_path: Path):
    with pytest.raises(NgioValueError):
        create_ome_zarr_from_array(
            array=np.random.randint(0, 255, (64, 64), dtype="uint8"),
            store=tmp_path / "fail.zarr",
            xy_pixelsize=0.5,
            axes_names=["z", "y", "x"],  # should fail expected yx
            levels=1,
            overwrite=True,
        )

    with pytest.raises(NgioValueError):
        create_ome_zarr_from_array(
            array=np.random.randint(0, 255, (2, 64, 64), dtype="uint8"),
            store=tmp_path / "fail.zarr",
            xy_pixelsize=0.5,
            axes_names=["c", "y", "x"],
            levels=1,
            channel_labels=[
                "channel1",
                "channel2",
                "channel3",
            ],  # should fail expected 2 channels
            overwrite=True,
        )

    with pytest.raises(NgioValueError):
        create_ome_zarr_from_array(
            array=np.random.randint(0, 255, (2, 64, 64), dtype="uint8"),
            store=tmp_path / "fail.zarr",
            xy_pixelsize=0.5,
            axes_names=["c", "y", "x"],
            levels=1,
            chunks=(1, 64, 64, 64),  # should fail expected 3 axes
            overwrite=True,
        )



"""Various utilities for the ngio package."""

import os

from ngio.common._common_types import ArrayLike
from ngio.utils._datasets import download_ome_zarr_dataset, list_ome_zarr_datasets
from ngio.utils._errors import (
    NgioFileExistsError,
    NgioFileNotFoundError,
    NgioTableValidationError,
    NgioValidationError,
    NgioValueError,
)
from ngio.utils._fractal_fsspec_store import fractal_fsspec_store
from ngio.utils._logger import ngio_logger, set_logger_level
from ngio.utils._zarr_utils import (
    AccessModeLiteral,
    StoreOrGroup,
    ZarrGroupHandler,
    open_group_wrapper,
)

set_logger_level(os.getenv("NGIO_LOGGER_LEVEL", "WARNING"))

__all__ = [
    # Zarr
    "AccessModeLiteral",
    "ArrayLike",
    # Errors
    "NgioFileExistsError",
    "NgioFileNotFoundError",
    "NgioTableValidationError",
    "NgioValidationError",
    "NgioValueError",
    "StoreOrGroup",
    "ZarrGroupHandler",
    # Datasets
    "download_ome_zarr_dataset",
    # Fractal
    "fractal_fsspec_store",
    "list_ome_zarr_datasets",
    # Logger
    "ngio_logger",
    "open_group_wrapper",
    "set_logger_level",
]



"""Utility functions for working with OME-Zarr images."""

from collections.abc import Collection
from typing import TypeVar

from ngio.common._pyramid import init_empty_pyramid
from ngio.ome_zarr_meta import (
    NgioImageMeta,
    NgioLabelMeta,
    PixelSize,
    get_image_meta_handler,
    get_label_meta_handler,
)
from ngio.ome_zarr_meta.ngio_specs import (
    DefaultNgffVersion,
    DefaultSpaceUnit,
    DefaultTimeUnit,
    NgffVersions,
    SpaceUnits,
    TimeUnits,
    canonical_axes_order,
    canonical_label_axes_order,
)
from ngio.utils import NgioValueError, StoreOrGroup, ZarrGroupHandler

_image_or_label_meta = TypeVar("_image_or_label_meta", NgioImageMeta, NgioLabelMeta)


def _init_generic_meta(
    meta_type: type[_image_or_label_meta],
    pixelsize: float,
    axes_names: Collection[str],
    z_spacing: float = 1.0,
    time_spacing: float = 1.0,
    levels: int | list[str] = 5,
    yx_scaling_factor: float | tuple[float, float] = 2.0,
    z_scaling_factor: float = 1.0,
    space_unit: SpaceUnits | str | None = DefaultSpaceUnit,
    time_unit: TimeUnits | str | None = DefaultTimeUnit,
    name: str | None = None,
    version: NgffVersions = DefaultNgffVersion,
) -> tuple[_image_or_label_meta, list[float]]:
    """Initialize the metadata for an image or label."""
    scaling_factors = []
    for ax in axes_names:
        if ax == "z":
            scaling_factors.append(z_scaling_factor)
        elif ax in ["x"]:
            if isinstance(yx_scaling_factor, tuple):
                scaling_factors.append(yx_scaling_factor[1])
            else:
                scaling_factors.append(yx_scaling_factor)
        elif ax in ["y"]:
            if isinstance(yx_scaling_factor, tuple):
                scaling_factors.append(yx_scaling_factor[0])
            else:
                scaling_factors.append(yx_scaling_factor)
        else:
            scaling_factors.append(1.0)

    pixel_sizes = PixelSize(
        x=pixelsize,
        y=pixelsize,
        z=z_spacing,
        t=time_spacing,
        space_unit=space_unit,
        time_unit=time_unit,
    )

    meta = meta_type.default_init(
        name=name,
        levels=levels,
        axes_names=axes_names,
        pixel_size=pixel_sizes,
        scaling_factors=scaling_factors,
        version=version,
    )
    return meta, scaling_factors


def create_empty_label_container(
    store: StoreOrGroup,
    shape: Collection[int],
    pixelsize: float,
    z_spacing: float = 1.0,
    time_spacing: float = 1.0,
    levels: int | list[str] = 5,
    yx_scaling_factor: float | tuple[float, float] = 2.0,
    z_scaling_factor: float = 1.0,
    space_unit: SpaceUnits | str | None = DefaultSpaceUnit,
    time_unit: TimeUnits | str | None = DefaultTimeUnit,
    axes_names: Collection[str] | None = None,
    name: str | None = None,
    chunks: Collection[int] | None = None,
    dtype: str = "uint16",
    overwrite: bool = False,
    version: NgffVersions = DefaultNgffVersion,
) -> ZarrGroupHandler:
    """Create an empty label with the given shape and metadata.

    Args:
        store (StoreOrGroup): The Zarr store or group to create the image in.
        shape (Collection[int]): The shape of the image.
        pixelsize (float): The pixel size in x and y dimensions.
        z_spacing (float, optional): The spacing between z slices. Defaults to 1.0.
        time_spacing (float, optional): The spacing between time points.
            Defaults to 1.0.
        levels (int | list[str], optional): The number of levels in the pyramid or a
            list of level names. Defaults to 5.
        yx_scaling_factor (float, optional): The down-scaling factor in x and y
            dimensions. Defaults to 2.0.
        z_scaling_factor (float, optional): The down-scaling factor in z dimension.
            Defaults to 1.0.
        space_unit (SpaceUnits, optional): The unit of space. Defaults to
            DefaultSpaceUnit.
        time_unit (TimeUnits, optional): The unit of time. Defaults to
            DefaultTimeUnit.
        axes_names (Collection[str] | None, optional): The names of the axes.
            If None the canonical names are used. Defaults to None.
        name (str | None, optional): The name of the image. Defaults to None.
        chunks (Collection[int] | None, optional): The chunk shape. If None the shape
            is used. Defaults to None.
        dtype (str, optional): The data type of the image. Defaults to "uint16".
        overwrite (bool, optional): Whether to overwrite an existing image.
            Defaults to True.
        version (str, optional): The version of the OME-Zarr specification.
            Defaults to DefaultVersion.

    """
    if axes_names is None:
        axes_names = canonical_label_axes_order()[-len(shape) :]

    if len(axes_names) != len(shape):
        raise NgioValueError(
            f"Number of axes names {axes_names} does not match the number of "
            f"dimensions {shape}."
        )

    meta, scaling_factors = _init_generic_meta(
        meta_type=NgioLabelMeta,
        pixelsize=pixelsize,
        z_spacing=z_spacing,
        time_spacing=time_spacing,
        levels=levels,
        yx_scaling_factor=yx_scaling_factor,
        z_scaling_factor=z_scaling_factor,
        space_unit=space_unit,
        time_unit=time_unit,
        axes_names=axes_names,
        name=name,
        version=version,
    )

    mode = "w" if overwrite else "w-"
    group_handler = ZarrGroupHandler(store=store, mode=mode, cache=False)
    image_handler = get_label_meta_handler(version=version, group_handler=group_handler)
    image_handler.write_meta(meta)

    init_empty_pyramid(
        store=store,
        paths=meta.paths,
        scaling_factors=scaling_factors,
        ref_shape=shape,
        chunks=chunks,
        dtype=dtype,
        mode="a",
    )
    group_handler._mode = "r+"
    return group_handler


def create_empty_image_container(
    store: StoreOrGroup,
    shape: Collection[int],
    pixelsize: float,
    z_spacing: float = 1.0,
    time_spacing: float = 1.0,
    levels: int | list[str] = 5,
    yx_scaling_factor: float | tuple[float, float] = 2,
    z_scaling_factor: float = 1.0,
    space_unit: SpaceUnits | str | None = DefaultSpaceUnit,
    time_unit: TimeUnits | str | None = DefaultTimeUnit,
    axes_names: Collection[str] | None = None,
    name: str | None = None,
    chunks: Collection[int] | None = None,
    dtype: str = "uint16",
    overwrite: bool = False,
    version: NgffVersions = DefaultNgffVersion,
) -> ZarrGroupHandler:
    """Create an empty OME-Zarr image with the given shape and metadata.

    Args:
        store (StoreOrGroup): The Zarr store or group to create the image in.
        shape (Collection[int]): The shape of the image.
        pixelsize (float): The pixel size in x and y dimensions.
        z_spacing (float, optional): The spacing between z slices. Defaults to 1.0.
        time_spacing (float, optional): The spacing between time points.
            Defaults to 1.0.
        levels (int | list[str], optional): The number of levels in the pyramid or a
            list of level names. Defaults to 5.
        yx_scaling_factor (float, optional): The down-scaling factor in x and y
            dimensions. Defaults to 2.0.
        z_scaling_factor (float, optional): The down-scaling factor in z dimension.
            Defaults to 1.0.
        space_unit (SpaceUnits, optional): The unit of space. Defaults to
            DefaultSpaceUnit.
        time_unit (TimeUnits, optional): The unit of time. Defaults to
            DefaultTimeUnit.
        axes_names (Collection[str] | None, optional): The names of the axes.
            If None the canonical names are used. Defaults to None.
        name (str | None, optional): The name of the image. Defaults to None.
        chunks (Collection[int] | None, optional): The chunk shape. If None the shape
            is used. Defaults to None.
        dtype (str, optional): The data type of the image. Defaults to "uint16".
        overwrite (bool, optional): Whether to overwrite an existing image.
            Defaults to True.
        version (str, optional): The version of the OME-Zarr specification.
            Defaults to DefaultVersion.

    """
    if axes_names is None:
        axes_names = canonical_axes_order()[-len(shape) :]

    if len(axes_names) != len(shape):
        raise NgioValueError(
            f"Number of axes names {axes_names} does not match the number of "
            f"dimensions {shape}."
        )

    meta, scaling_factors = _init_generic_meta(
        meta_type=NgioImageMeta,
        pixelsize=pixelsize,
        z_spacing=z_spacing,
        time_spacing=time_spacing,
        levels=levels,
        yx_scaling_factor=yx_scaling_factor,
        z_scaling_factor=z_scaling_factor,
        space_unit=space_unit,
        time_unit=time_unit,
        axes_names=axes_names,
        name=name,
        version=version,
    )
    mode = "w" if overwrite else "w-"
    group_handler = ZarrGroupHandler(store=store, mode=mode, cache=False)
    image_handler = get_image_meta_handler(version=version, group_handler=group_handler)
    image_handler.write_meta(meta)

    init_empty_pyramid(
        store=store,
        paths=meta.paths,
        scaling_factors=scaling_factors,
        ref_shape=shape,
        chunks=chunks,
        dtype=dtype,
        mode="a",
    )

    group_handler._mode = "r+"
    return group_handler



"""Ngio Tables implementations."""

from ngio.tables.backends import ImplementedTableBackends
from ngio.tables.tables_container import (
    FeatureTable,
    GenericRoiTable,
    MaskingRoiTable,
    RoiTable,
    Table,
    TablesContainer,
    TypedTable,
    open_table,
    open_tables_container,
)
from ngio.tables.v1._generic_table import GenericTable

__all__ = [
    "FeatureTable",
    "GenericRoiTable",
    "GenericTable",
    "ImplementedTableBackends",
    "MaskingRoiTable",
    "RoiTable",
    "Table",
    "TablesContainer",
    "TypedTable",
    "open_table",
    "open_tables_container",
]



"""OME-Zarr object models."""

from ngio.images.image import Image, ImagesContainer
from ngio.images.label import Label, LabelsContainer
from ngio.images.ome_zarr_container import (
    OmeZarrContainer,
    create_empty_ome_zarr,
    create_ome_zarr_from_array,
    open_image,
    open_ome_zarr_container,
)

__all__ = [
    "Image",
    "ImagesContainer",
    "Label",
    "LabelsContainer",
    "OmeZarrContainer",
    "create_empty_ome_zarr",
    "create_ome_zarr_from_array",
    "open_image",
    "open_ome_zarr_container",
]
